"""
    tests.test_openapi
    ------------------

    Tests some stuff of ``sphinxcontrib.openapi`` module.

    :copyright: (c) 2016, Ihor Kalnytskyi.
    :license: BSD, see LICENSE for details.
"""

import os
import textwrap
import collections

import py
import pytest

import yaml

from sphinxcontrib.openapi import renderers
from sphinxcontrib.openapi import openapi20
from sphinxcontrib.openapi import utils


class TestOpenApi2HttpDomain(object):

    def test_basic(self):
        renderer = renderers.HttpdomainOldRenderer(None, {})
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'paths': {
                '/resources/{kind}': {
                    'get': {
                        'summary': 'List Resources',
                        'description': '~ some useful description ~',
                        'parameters': [
                            {
                                'name': 'kind',
                                'in': 'path',
                                'type': 'string',
                                'description': 'Kind of resource to list.',
                            },
                            {
                                'name': 'limit',
                                'in': 'query',
                                'type': 'integer',
                                'description': 'Show up to `limit` entries.',
                            },
                            {
                                'name': 'If-None-Match',
                                'in': 'header',
                                'type': 'string',
                                'description': 'Last known resource ETag.'
                            },
                        ],
                        'responses': {
                            '200': {
                                'description': 'An array of resources.',
                                'headers': {
                                    'ETag': {
                                        'description': 'Resource ETag.',
                                        'type': 'string'
                                    },
                                },
                            },
                        },
                    },
                },
            },
        }))

        assert text == textwrap.dedent('''
            .. http:get:: /resources/{kind}
               :synopsis: List Resources

               **List Resources**

               ~ some useful description ~

               :param string kind:
                  Kind of resource to list.
               :query integer limit:
                  Show up to `limit` entries.
               :status 200:
                  An array of resources.
               :reqheader If-None-Match:
                  Last known resource ETag.
               :resheader ETag:
                  Resource ETag.
        ''').lstrip()

    def test_groups(self):
        renderer = renderers.HttpdomainOldRenderer(None, {'group': True})
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'tags': [
                {'name': 'tags'},
                {'name': 'pets'},
            ],
            'paths': collections.OrderedDict([
                ('/', {
                    'get': {
                        'summary': 'Index',
                        'description': '~ some useful description ~',
                        'responses': {
                            '200': {
                                'description': 'Index',
                                'content': {
                                    'application/json': {
                                        'pets': 'https://example.com/api/pets',
                                        'tags': 'https://example.com/api/tags',
                                    }
                                }
                            },
                        },
                    },
                }),
                ('/pets', {
                    'get': {
                        'summary': 'List Pets',
                        'description': '~ some useful description ~',
                        'responses': {
                            '200': {
                                'description': 'Pets',
                                'content': {
                                    'application/json': [
                                        {
                                            'example': '{"foo": "bar"}'
                                        },
                                    ],
                                },
                            },
                        },
                        'tags': [
                            'pets',
                        ],
                    },
                }),
                ('/pets/{name}', {
                    'get': {
                        'summary': 'Show Pet',
                        'description': '~ some useful description ~',
                        'parameters': [
                            {
                                'name': 'name',
                                'in': 'path',
                                'type': 'string',
                                'description': 'Name of pet.',
                            },
                        ],
                        'responses': {
                            '200': {
                                'description': 'A Pet',
                                'content': {
                                    'application/json': {
                                        'example': '{"foo": "bar"}'
                                    }
                                }
                            },
                        },
                        'tags': [
                            'pets',
                        ],
                    },
                }),
                ('/tags', {
                    'get': {
                        'summary': 'List Tags',
                        'description': '~ some useful description ~',
                        'responses': {
                            '200': {
                                'description': 'Tags',
                                'content': {
                                    'application/json': [
                                        {
                                            'example': '{"foo": "bar"}'
                                        },
                                    ],
                                }
                            },
                        },
                        'tags': [
                            'tags',
                            'pets',
                        ],
                    },
                }),
            ]),
        }))
        assert text == textwrap.dedent('''
            tags
            ====

            .. http:get:: /tags
               :synopsis: List Tags

               **List Tags**

               ~ some useful description ~

               :status 200:
                  Tags

            pets
            ====

            .. http:get:: /pets
               :synopsis: List Pets

               **List Pets**

               ~ some useful description ~

               :status 200:
                  Pets

            .. http:get:: /pets/{name}
               :synopsis: Show Pet

               **Show Pet**

               ~ some useful description ~

               :param string name:
                  Name of pet.
               :status 200:
                  A Pet

            default
            =======

            .. http:get:: /
               :synopsis: Index

               **Index**

               ~ some useful description ~

               :status 200:
                  Index
        ''').lstrip()

    def test_two_resources(self):
        spec = collections.defaultdict(collections.OrderedDict)
        spec['paths']['/resource_a'] = {
            'get': {
                'description': 'resource a',
                'responses': {
                    '200': {'description': 'ok'},
                }
            }
        }
        spec['paths']['/resource_b'] = {
            'post': {
                'description': 'resource b',
                'responses': {
                    '404': {'description': 'error'},
                }
            }
        }

        renderer = renderers.HttpdomainOldRenderer(None, {})
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent('''
            .. http:get:: /resource_a
               :synopsis: null

               resource a

               :status 200:
                  ok

            .. http:post:: /resource_b
               :synopsis: null

               resource b

               :status 404:
                  error
        ''').lstrip()

    def test_path_option(self):
        spec = collections.defaultdict(collections.OrderedDict)
        spec['paths']['/resource_a'] = {
            'get': {
                'description': 'resource a',
                'responses': {
                    '200': {'description': 'ok'},
                }
            }
        }
        spec['paths']['/resource_b'] = {
            'post': {
                'description': 'resource b',
                'responses': {
                    '404': {'description': 'error'},
                }
            }
        }

        renderer = renderers.HttpdomainOldRenderer(None, {'paths': [
            '/resource_a',
        ]})
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent('''
            .. http:get:: /resource_a
               :synopsis: null

               resource a

               :status 200:
                  ok
        ''').lstrip()

    def test_include_option(self):
        spec = collections.defaultdict(collections.OrderedDict)
        spec['paths']['/resource_a'] = {
            'get': {
                'description': 'resource a',
                'responses': {
                    '200': {'description': 'ok'},
                }
            }
        }
        spec['paths']['/resource_b'] = {
            'post': {
                'description': 'resource b',
                'responses': {
                    '404': {'description': 'error'},
                }
            }
        }

        renderer = renderers.HttpdomainOldRenderer(None, {'include': [
            '/resource',
        ]})
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent('''
            .. http:get:: /resource_a
               :synopsis: null

               resource a

               :status 200:
                  ok

            .. http:post:: /resource_b
               :synopsis: null

               resource b

               :status 404:
                  error
        ''').lstrip()

    def test_exclude_option(self):
        spec = collections.defaultdict(collections.OrderedDict)
        spec['paths']['/resource_a'] = {
            'get': {
                'description': 'resource a',
                'responses': {
                    '200': {'description': 'ok'},
                }
            }
        }
        spec['paths']['/resource_b'] = {
            'post': {
                'description': 'resource b',
                'responses': {
                    '404': {'description': 'error'},
                }
            }
        }

        renderer = renderers.HttpdomainOldRenderer(None, {'exclude': [
            '/.*_a',
        ]})
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent('''
            .. http:post:: /resource_b
               :synopsis: null

               resource b

               :status 404:
                  error
        ''').lstrip()

    def test_method_option(self):
        spec = collections.defaultdict(collections.OrderedDict)
        spec['paths']['/resource_a'] = {
            'get': {
                'description': 'resource a',
                'responses': {
                    '200': {'description': 'ok'},
                }
            },
            'post': {
                'description': 'resource a',
                'responses': {
                    '201': {'description': 'ok'},
                }
            },
            'put': {
                'description': 'resource a',
                'responses': {
                    '404': {'description': 'error'},
                }
            }
        }

        renderer = renderers.HttpdomainOldRenderer(
            None,
            {
                'methods': ['post'],
                'paths': ['/resource_a'],
            },
        )
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))

        assert text == textwrap.dedent('''
            .. http:post:: /resource_a
               :synopsis: null

               resource a

               :status 201:
                  ok
        ''').lstrip()

    def test_root_parameters(self):
        spec = {'paths': {}}
        spec['paths']['/resources/{name}'] = collections.OrderedDict()

        spec['paths']['/resources/{name}']['parameters'] = [
            {
                'name': 'name',
                'in': 'path',
                'type': 'string',
                'description': 'The name of the resource.',
            }
        ]
        spec['paths']['/resources/{name}']['get'] = {
            'summary': 'Fetch a Resource',
            'description': '~ some useful description ~',
            'responses': {
                '200': {
                    'description': 'The fetched resource.',
                },
            },
        }
        spec['paths']['/resources/{name}']['put'] = {
            'summary': 'Modify a Resource',
            'description': '~ some useful description ~',
            'responses': {
                '200': {
                    'description': 'The modified resource.',
                },
            },
        }

        renderer = renderers.HttpdomainOldRenderer(None, {})
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))

        assert text == textwrap.dedent('''
            .. http:get:: /resources/{name}
               :synopsis: Fetch a Resource

               **Fetch a Resource**

               ~ some useful description ~

               :param string name:
                  The name of the resource.
               :status 200:
                  The fetched resource.

            .. http:put:: /resources/{name}
               :synopsis: Modify a Resource

               **Modify a Resource**

               ~ some useful description ~

               :param string name:
                  The name of the resource.
               :status 200:
                  The modified resource.
        ''').lstrip()

    def test_path_invalid(self):
        spec = collections.defaultdict(collections.OrderedDict)
        spec['paths']['/resource_a'] = {
            'get': {
                'description': 'resource a',
                'responses': {
                    '200': {'description': 'ok'},
                }
            }
        }
        spec['paths']['/resource_b'] = {
            'post': {
                'description': 'resource b',
                'responses': {
                    '404': {'description': 'error'},
                }
            }
        }

        renderer = renderers.HttpdomainOldRenderer(None, {'paths': [
            '/resource_a',
            '/resource_invalid_name',
        ]})

        with pytest.raises(ValueError) as exc:
            '\n'.join(renderer.render_restructuredtext_markup(spec))

        assert str(exc.value) == (
            'One or more paths are not defined in the spec: '
            '/resource_invalid_name.'
        )

    def test_unicode_is_allowed(self):
        spec = {
            'paths': {
                '/resource_a': {
                    'get': {
                        'description': '\u041f',
                        'responses': {
                            '200': {'description': 'ok'}
                        }
                    }
                }
            }
        }

        renderer = renderers.HttpdomainOldRenderer(None, {})
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))

        assert text == textwrap.dedent('''
            .. http:get:: /resource_a
               :synopsis: null

               \u041f

               :status 200:
                  ok
        ''').lstrip()

    def test_json_in_out(self):
        renderer = renderers.HttpdomainOldRenderer(None, {})
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'definitions': {
                'CreateResourceSchema': {
                    'additionalProperties': False,
                    'properties': {
                        'string_field': {
                            'type': 'string',
                            'description': 'some input string'
                        },
                        'int_field': {
                            'default': 1,
                            'type': 'integer',
                        },
                    },
                    'required': [
                        'string_field'
                    ],
                    'title': 'CreateResourceSchema',
                    'type': 'object'
                },
                'ResourceSchema': {
                    'properties': {
                        'string_field': {
                            'type': 'string',
                            'description': 'some output string'
                        },
                        'int_field': {
                            'type': 'integer',
                        },
                    },
                    'required': [
                        'string_field'
                    ],
                    'title': 'ResourceSchema',
                    'type': 'object'
                },
                'Error': {
                    'properties': {
                        'errors': {
                            'type': 'object'
                        },
                        'message': {
                            'type': 'string'
                        }
                    },
                    'required': [
                        'message'
                    ],
                    'title': 'Error',
                    'type': 'object'
                },
            },
            'paths': {
                '/resources': {
                    'post': {
                        'description': '~ some useful description ~',
                        'parameters': [
                            {
                                'in': 'body',
                                'name': 'CreateResourceSchema',
                                'required': True,
                                'schema': {
                                    '$ref':
                                        '#/definitions/CreateResourceSchema'
                                }
                            },
                        ],
                        'responses': {
                            '201': {
                                'description': '~ some useful description ~',
                                'schema': {
                                    '$ref': '#/definitions/ResourceSchema'
                                }
                            },
                            'default': {
                                'description': '~ some useful description ~',
                                'schema': {
                                    '$ref': '#/definitions/Error'
                                }
                            }
                        },
                    },
                },
            },
        }))

        text2 = textwrap.dedent('''
            .. http:post:: /resources
               :synopsis: null

               ~ some useful description ~


               :<json integer int_field:
               :<json string string_field: some input string (required)

               :status 201:
                  ~ some useful description ~
               :status default:
                  ~ some useful description ~

               :>json integer int_field:
               :>json string string_field: some output string (required)

        ''').lstrip()
        assert text == text2


class TestOpenApi3HttpDomain(object):

    def test_basic(self):
        renderer = renderers.HttpdomainOldRenderer(
            None,
            {
                'examples': True,
                'group_examples': True
            }
        )
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'openapi': '3.0.0',
            'paths': {
                '/resources/{kind}': {
                    'get': {
                        'summary': 'List Resources',
                        'description': '~ some useful description ~',
                        'parameters': [
                            {
                                'name': 'kind',
                                'in': 'path',
                                'schema': {'type': 'string'},
                                'description': 'Kind of resource to list.',
                            },
                            {
                                'name': 'limit',
                                'in': 'query',
                                'schema': {'type': 'integer'},
                                'description': 'Show up to `limit` entries.',
                                'required': True,
                            },
                            {
                                'name': 'offset',
                                'in': 'query',
                                'schema': {'type': 'integer'},
                                'description': 'Start from offset.',
                            },
                            {
                                'name': 'If-None-Match',
                                'in': 'header',
                                'schema': {'type': 'string'},
                                'description': 'Last known resource ETag.'
                            },
                        ],
                        'requestBody': {
                            'content': {
                                'application/json':  {
                                    'example': '{"foo2": "bar2"}'
                                }
                            }
                        },
                        'responses': {
                            '200': {
                                'description': 'An array of resources.',
                                'content': {
                                    'application/json': {
                                        'example': '{"foo": "bar"}'
                                    }
                                }
                            },
                        },
                    },
                },
            },
        }))
        assert text == textwrap.dedent('''
            .. http:get:: /resources/{kind}
               :synopsis: List Resources

               **List Resources**

               ~ some useful description ~

               :param string kind:
                  Kind of resource to list.
               :query integer limit:
                  Show up to `limit` entries.
                  (Required)
               :query integer offset:
                  Start from offset.
               :reqheader If-None-Match:
                  Last known resource ETag.
               :status 200:
                  An array of resources.

               **Example request:**

               .. sourcecode:: http

                  GET /resources/{kind}?limit=1&offset=1 HTTP/1.1
                  Host: example.com
                  Content-Type: application/json

                  {"foo2": "bar2"}


               **Example request:**

               .. sourcecode:: http

                  GET /resources/{kind}?limit=1&offset=1 HTTP/1.1
                  Host: example.com


               **Example response:**

               .. sourcecode:: http

                  HTTP/1.1 200 OK
                  Content-Type: application/json

                  {"foo": "bar"}

        ''').lstrip()

    def test_rfc7807(self):
        # Fix order to have a reliable test
        pb_example = collections.OrderedDict()
        pb_example["type"] = "string"
        pb_example["title"] = "string"
        pb_example["status"] = 1
        pb_example["detail"] = "string"
        pb_example["instance"] = "string"
        renderer = renderers.HttpdomainOldRenderer(None, {'examples': True})
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'openapi': '3.0.0',
            'paths': {
                '/problem': {
                    'post': {
                        'summary': 'Problem',
                        'description': '~ some useful description ~',
                        'requestBody': {
                            'content': {
                                'application/problem+json':  {
                                    'example': pb_example
                                }
                            }
                        },
                        'responses': {
                            '200': {
                                'description': 'An array of resources.',
                                'content': {
                                    'application/json': {
                                        'example': {"foo": "bar"}
                                    }
                                }
                            },
                        },
                    },
                },
            },
        }))
        assert text == textwrap.dedent('''
            .. http:post:: /problem
               :synopsis: Problem

               **Problem**

               ~ some useful description ~


               **Example request:**

               .. sourcecode:: http

                  POST /problem HTTP/1.1
                  Host: example.com
                  Content-Type: application/problem+json

                  {
                      "type": "string",
                      "title": "string",
                      "status": 1,
                      "detail": "string",
                      "instance": "string"
                  }

               :status 200:
                  An array of resources.

                  **Example response:**

                  .. sourcecode:: http

                     HTTP/1.1 200 OK
                     Content-Type: application/json

                     {
                         "foo": "bar"
                     }

        ''').lstrip()

    def test_group_examples(self):
        renderer = renderers.HttpdomainOldRenderer(None, {
            'examples': True,
            'group_examples': True
            }
        )
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'openapi': '3.0.0',
            'paths': {
                '/problem': {
                    'post': {
                        'summary': 'Problem',
                        'description': '~ some useful description ~',
                        'requestBody': {
                            'content': {
                                'application/json':  {
                                    'example': {"foo": "bar"}
                                }
                            }
                        },
                        'responses': {
                            '200': {
                                'description': 'An array of resources.',
                                'content': {
                                    'application/json': {
                                        'example': {"foo": "bar"}
                                    }
                                }
                            },
                        },
                    },
                },
            },
        }))
        assert text == textwrap.dedent('''
            .. http:post:: /problem
               :synopsis: Problem

               **Problem**

               ~ some useful description ~

               :status 200:
                  An array of resources.

               **Example request:**

               .. sourcecode:: http

                  POST /problem HTTP/1.1
                  Host: example.com
                  Content-Type: application/json

                  {
                      "foo": "bar"
                  }


               **Example response:**

               .. sourcecode:: http

                  HTTP/1.1 200 OK
                  Content-Type: application/json

                  {
                      "foo": "bar"
                  }

        ''').lstrip()

    def test_groups(self):
        renderer = renderers.HttpdomainOldRenderer(None, {'group': True})
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'openapi': '3.0.0',
            'tags': [
                {'name': 'tags'},
                {'name': 'pets'},
            ],
            'paths': collections.OrderedDict([
                ('/', {
                    'get': {
                        'summary': 'Index',
                        'description': '~ some useful description ~',
                        'deprecated': True,
                        'responses': {
                            '200': {
                                'description': 'Index',
                                'content': {
                                    'application/json': {
                                        'pets': 'https://example.com/api/pets',
                                        'tags': 'https://example.com/api/tags',
                                    }
                                }
                            },
                        },
                    },
                }),
                ('/pets', {
                    'get': {
                        'summary': 'List Pets',
                        'description': '~ some useful description ~',
                        'responses': {
                            '200': {
                                'description': 'Pets',
                                'content': {
                                    'application/json': [
                                        {
                                            'example': '{"foo": "bar"}'
                                        },
                                    ],
                                },
                            },
                        },
                        'tags': [
                            'pets',
                        ],
                    },
                }),
                ('/pets/{name}', {
                    'get': {
                        'summary': 'Show Pet',
                        'description': '~ some useful description ~',
                        'parameters': [
                            {
                                'name': 'name',
                                'in': 'path',
                                'schema': {'type': 'string'},
                                'description': 'Name of pet.',
                            },
                        ],
                        'responses': {
                            '200': {
                                'description': 'A Pet',
                                'content': {
                                    'application/json': {
                                        'example': '{"foo": "bar"}'
                                    }
                                }
                            },
                        },
                        'tags': [
                            'pets',
                        ],
                    },
                }),
                ('/tags', {
                    'get': {
                        'summary': 'List Tags',
                        'description': '~ some useful description ~',
                        'responses': {
                            '200': {
                                'description': 'Tags',
                                'content': {
                                    'application/json': [
                                        {
                                            'example': '{"foo": "bar"}'
                                        },
                                    ],
                                }
                            },
                        },
                        'tags': [
                            'tags',
                            'pets',
                        ],
                    },
                }),
            ]),
        }))
        assert text == textwrap.dedent('''
            tags
            ====

            .. http:get:: /tags
               :synopsis: List Tags

               **List Tags**

               ~ some useful description ~

               :status 200:
                  Tags.

            pets
            ====

            .. http:get:: /pets
               :synopsis: List Pets

               **List Pets**

               ~ some useful description ~

               :status 200:
                  Pets.

            .. http:get:: /pets/{name}
               :synopsis: Show Pet

               **Show Pet**

               ~ some useful description ~

               :param string name:
                  Name of pet.
               :status 200:
                  A Pet.

            default
            =======

            .. http:get:: /
               :synopsis: Index

               **Index**

               ~ some useful description ~

               **DEPRECATED**

               :status 200:
                  Index.
        ''').lstrip()

    def test_required_parameters(self):
        renderer = renderers.HttpdomainOldRenderer(None, {})
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'openapi': '3.0.0',
            'paths': {
                '/resources/{kind}': {
                    'get': {
                        'summary': 'List Resources',
                        'description': '~ some useful description ~',
                        'parameters': [
                            {
                                'name': 'kind',
                                'in': 'path',
                                'schema': {'type': 'string'},
                                'description': 'Kind of resource to list.',
                            },
                            {
                                'name': 'limit',
                                'in': 'query',
                                'required': True,
                                'schema': {'type': 'integer'},
                                'description': 'Show up to `limit` entries.',
                            },
                            {
                                'name': 'If-None-Match',
                                'in': 'header',
                                'required': True,
                                'schema': {'type': 'string'},
                                'description': 'Last known resource ETag.'
                            },
                        ],
                        'requestBody': {
                            'content': {
                                'application/json':  {
                                    'example': '{"foo2": "bar2"}'
                                }
                            }
                        },
                        'responses': {
                            '200': {
                                'description': 'An array of resources.',
                                'content': {
                                    'application/json': {
                                        'example': '{"foo": "bar"}'
                                    }
                                }
                            },
                        },
                    },
                },
            },
        }))
        assert text == textwrap.dedent('''
            .. http:get:: /resources/{kind}
               :synopsis: List Resources

               **List Resources**

               ~ some useful description ~

               :param string kind:
                  Kind of resource to list.
               :query integer limit:
                  Show up to `limit` entries.
                  (Required)
               :reqheader If-None-Match:
                  Last known resource ETag.
                  (Required)
               :status 200:
                  An array of resources.
        ''').lstrip()

    def test_example_generation(self):
        renderer = renderers.HttpdomainOldRenderer(None, {'examples': True})
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'openapi': '3.0.0',
            'paths': collections.OrderedDict([
                ('/resources/', collections.OrderedDict([
                    ('get', {
                        'summary': 'List Resources',
                        'description': '~ some useful description ~',
                        'parameters': [
                            {
                                'name': 'kind',
                                'in': 'path',
                                'schema': {'type': 'string'},
                                'description': 'Kind of resource to list.',
                            },
                            {
                                'name': 'limit',
                                'in': 'query',
                                'required': True,
                                'schema': {'type': 'integer'},
                                'description': 'Show up to `limit` entries.',
                            },
                            {
                                'name': 'If-None-Match',
                                'in': 'header',
                                'schema': {'type': 'string'},
                                'description': 'Last known resource ETag.'
                            },
                        ],
                        'responses': {
                            '200': {
                                'description': 'An array of resources.',
                                'content': {
                                    'application/json': {
                                        'schema': {
                                            'type': 'array',
                                            'items': {
                                                '$ref': '#/components/schemas/Resource',  # noqa
                                            },
                                        },
                                    }
                                }
                            },
                        },
                    }),
                    ('post', {
                        'summary': 'Create Resource',
                        'description': '~ some useful description ~',
                        'parameters': [],
                        'requestBody': {
                            'content': {
                                'application/json':  {
                                    'schema': {
                                            'type': 'array',
                                            'items': {
                                                '$ref': '#/components/schemas/Resource',  # noqa
                                            },
                                    },
                                }
                            }
                        },
                        'responses': {
                            '200': {
                                'description': 'The created resource.',
                                'content': {
                                    'application/json': {
                                        'schema': {
                                            '$ref': '#/components/schemas/Resource',  # noqa
                                        },
                                    }
                                }
                            },
                        },
                    }),
                ])),
                ('/resources/{kind}', collections.OrderedDict([
                    ('get', {
                        'summary': 'Show Resource',
                        'description': '~ some useful description ~',
                        'parameters': [
                            {
                                'name': 'kind',
                                'in': 'path',
                                'schema': {'type': 'string'},
                                'description': 'Kind of resource to list.',
                            },
                        ],
                        'responses': {
                            '200': {
                                'description': 'The created resource.',
                                'content': {
                                    'application/json': {
                                        'schema': {
                                            '$ref': '#/components/schemas/Resource',  # noqa
                                        },
                                    }
                                }
                            },
                        },
                    }),
                    ('patch', {
                        'summary': 'Update Resource (partial)',
                        'description': '~ some useful description ~',
                        'parameters': [
                            {
                                'name': 'kind',
                                'in': 'path',
                                'schema': {'type': 'string'},
                                'description': 'Kind of resource to list.',
                            },
                        ],
                        'requestBody': {
                            'content': {
                                'application/json':  {
                                    'schema': {
                                        '$ref': '#/components/schemas/Resource',  # noqa
                                    },
                                }
                            }
                        },
                        'responses': {
                            '200': {
                                'description': 'The created resource.',
                                'content': {
                                    'application/json': {
                                        'schema': {
                                            '$ref': '#/components/schemas/Resource',  # noqa
                                        },
                                    }
                                }
                            },
                        },
                    }),
                ])),
            ]),
            'components': {
                'schemas': {
                    'Resource': {
                        'type': 'object',
                        'properties': collections.OrderedDict([
                            ('kind', {
                                'title': 'Kind',
                                'type': 'string',
                                'readOnly': True,
                            }),
                            ('description', {
                                'title': 'Description',
                                'type': 'string',
                            }),
                            ('data', {
                                'title': 'Data',
                                'type': 'string',
                                'format': 'byte',
                            }),
                        ]),
                        'example': {
                            'kind': 'KIND',
                            'description': 'DESC',
                            'data': 'ABCD',
                        }
                    },
                },
            },
        }))

        assert text == textwrap.dedent('''
            .. http:get:: /resources/
               :synopsis: List Resources

               **List Resources**

               ~ some useful description ~

               :param string kind:
                  Kind of resource to list.
               :query integer limit:
                  Show up to `limit` entries.
                  (Required)
               :reqheader If-None-Match:
                  Last known resource ETag.

               **Example request:**

               .. sourcecode:: http

                  GET /resources/?limit=1 HTTP/1.1
                  Host: example.com

               :status 200:
                  An array of resources.

                  **Example response:**

                  .. sourcecode:: http

                     HTTP/1.1 200 OK
                     Content-Type: application/json

                     [
                         {
                             "kind": "KIND",
                             "description": "DESC",
                             "data": "ABCD"
                         }
                     ]


            .. http:post:: /resources/
               :synopsis: Create Resource

               **Create Resource**

               ~ some useful description ~


               **Example request:**

               .. sourcecode:: http

                  POST /resources/ HTTP/1.1
                  Host: example.com
                  Content-Type: application/json

                  [
                      {
                          "description": "DESC",
                          "data": "ABCD"
                      }
                  ]

               :status 200:
                  The created resource.

                  **Example response:**

                  .. sourcecode:: http

                     HTTP/1.1 200 OK
                     Content-Type: application/json

                     {
                         "kind": "KIND",
                         "description": "DESC",
                         "data": "ABCD"
                     }


            .. http:get:: /resources/{kind}
               :synopsis: Show Resource

               **Show Resource**

               ~ some useful description ~

               :param string kind:
                  Kind of resource to list.

               **Example request:**

               .. sourcecode:: http

                  GET /resources/{kind} HTTP/1.1
                  Host: example.com

               :status 200:
                  The created resource.

                  **Example response:**

                  .. sourcecode:: http

                     HTTP/1.1 200 OK
                     Content-Type: application/json

                     {
                         "kind": "KIND",
                         "description": "DESC",
                         "data": "ABCD"
                     }


            .. http:patch:: /resources/{kind}
               :synopsis: Update Resource (partial)

               **Update Resource (partial)**

               ~ some useful description ~

               :param string kind:
                  Kind of resource to list.

               **Example request:**

               .. sourcecode:: http

                  PATCH /resources/{kind} HTTP/1.1
                  Host: example.com
                  Content-Type: application/json

                  {
                      "description": "DESC",
                      "data": "ABCD"
                  }

               :status 200:
                  The created resource.

                  **Example response:**

                  .. sourcecode:: http

                     HTTP/1.1 200 OK
                     Content-Type: application/json

                     {
                         "kind": "KIND",
                         "description": "DESC",
                         "data": "ABCD"
                     }

        ''').lstrip()

    def test_get_example_with_explode(self):
        renderer = renderers.HttpdomainOldRenderer(None, {'examples': True})
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'openapi': '3.0.0',
            'paths': collections.OrderedDict([
                ('/resources/', collections.OrderedDict([
                    ('get', {
                        'summary': 'List Resources',
                        'description': '~ some useful description ~',
                        'parameters': [
                            {
                                'name': 'params',
                                'in': 'query',
                                'required': True,
                                'schema': {
                                    'type': 'array',
                                    'items': {
                                        'type': 'string'
                                    }
                                },
                                'style': 'form',
                                'explode': True,
                                'example': [
                                    'p1',
                                    'p2',
                                ],
                                'description': 'List with explode set to True'
                            },
                            {
                                'name': 'values',
                                'in': 'query',
                                'required': True,
                                'schema': {
                                    'type': 'object',
                                    'additionalProperties': True
                                },
                                'style': 'form',
                                'explode': True,
                                'example': collections.OrderedDict([
                                    ('v1', 'V1'),
                                    ('v2', 'V2'),
                                ]),
                                'description': 'Dict with explode set to True'
                            },
                        ],
                        'responses': {
                            '200': {
                                'description': 'OK'
                            },
                        },
                    }),
                ])),
            ]),
        }))

        assert text == textwrap.dedent('''
            .. http:get:: /resources/
               :synopsis: List Resources

               **List Resources**

               ~ some useful description ~

               :query array params:
                  List with explode set to True.
                  (Required)
               :query object values:
                  Dict with explode set to True.
                  (Required)

               **Example request:**

               .. sourcecode:: http

                  GET /resources/?params=p1&params=p2&v1=V1&v2=V2 HTTP/1.1
                  Host: example.com

               :status 200:
                  OK.
        ''').lstrip()

    def test_callback(self):
        renderer = renderers.HttpdomainOldRenderer(None, {})
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'openapi': '3.0.0',
            'paths': {
                '/resources/{kind}': {
                    'post': {
                        'summary': 'List Resources',
                        'description': '~ some useful description ~',
                        'parameters': [
                            {
                                'name': 'kind',
                                'in': 'path',
                                'schema': {'type': 'string'},
                                'description': 'Kind of resource to list.',
                            },
                            {
                                'name': 'callback',
                                'in': 'query',
                                'description': 'the callback address',
                                'required': False,
                                'schema': {
                                    'type': 'string',
                                    'format': 'uri'
                                },
                                'example': 'http://client.com/callback'
                            }
                        ],
                        'requestBody': {
                            'content': {
                                'application/json':  {
                                    'example': '{"foo2": "bar2"}'
                                }
                            }
                        },
                        'responses': {
                            '202': {
                                'description': 'Something',
                                'content': {
                                    'application/json': {
                                        'example': '{"foo": "bar"}'
                                    }
                                }
                            },
                        },
                        'callbacks': {
                            'callback': {
                                '${request.query.callback}': {
                                    'post': {
                                        'summary': 'Response callback',
                                        'operationId': 'sampleCB',
                                        'requestBody': {
                                            'required': True,
                                            'description': 'Result',
                                            'content': {
                                                'application/json': {
                                                    'schema': {
                                                        'type': 'object',
                                                        'required': ['status'],
                                                        'properties': {
                                                            'status': {
                                                                'type':
                                                                    'string',
                                                                'enum': [
                                                                    'OK',
                                                                    'ERROR'
                                                                ]
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        },
                                        'responses': {
                                            '200': {
                                                'description': 'Success'
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                },
            },
        }))
        assert text == textwrap.dedent('''
            .. http:post:: /resources/{kind}
               :synopsis: List Resources

               **List Resources**

               ~ some useful description ~

               :param string kind:
                  Kind of resource to list.
               :query string callback:
                  the callback address.
               :status 202:
                  Something.

               .. admonition:: Callback: callback

                  .. http:post:: ${request.query.callback}
                     :synopsis: Response callback

                     **Response callback**

                     :jsonparam string status:
                     :status 200:
                        Success.

        ''').lstrip()

    def test_string_example(self):
        renderer = renderers.HttpdomainOldRenderer(None, {'examples': True})
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'openapi': '3.0.0',
            'paths': {
                '/resources': {
                    'get': {
                        'summary': 'Get resources',
                        'responses': {
                            '200': {
                                'description': 'Something',
                                'content': {
                                    'application/json': {
                                        'schema': {
                                            'type': 'string',
                                            'example': '"A sample"',
                                        }
                                    }
                                }
                            },
                        },
                    },
                },
            },
        }))

        assert text == textwrap.dedent('''
            .. http:get:: /resources
               :synopsis: Get resources

               **Get resources**


               **Example request:**

               .. sourcecode:: http

                  GET /resources HTTP/1.1
                  Host: example.com

               :status 200:
                  Something.

                  **Example response:**

                  .. sourcecode:: http

                     HTTP/1.1 200 OK
                     Content-Type: application/json

                     "A sample"

        ''').lstrip()

    def test_ref_example(self):
        renderer = renderers.HttpdomainOldRenderer(None, {'examples': True})
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'openapi': '3.0.0',
            'paths': {
                '/resources': {
                    'get': {
                        'summary': 'Get resources',
                        'responses': {
                            '200': {
                                'description': 'Something',
                                'content': {
                                    'application/json': {
                                        'schema': {
                                            '$ref':
                                                '#/components/schemas/Data',
                                        }
                                    }
                                }
                            },
                        },
                    },
                },
            },
            'components': {
                'schemas': {
                    'Data': {
                        'type': 'object',
                        'additionalProperties': True,
                        'example': {
                            'prop1': "Sample 1",
                        }
                    }
                }
            }
        }))

        assert text == textwrap.dedent('''
            .. http:get:: /resources
               :synopsis: Get resources

               **Get resources**


               **Example request:**

               .. sourcecode:: http

                  GET /resources HTTP/1.1
                  Host: example.com

               :status 200:
                  Something.

                  **Example response:**

                  .. sourcecode:: http

                     HTTP/1.1 200 OK
                     Content-Type: application/json

                     {
                         "prop1": "Sample 1",
                         "...": "..."
                     }

        ''').lstrip()

    def test_method_option(self):
        spec = collections.defaultdict(collections.OrderedDict)
        spec['paths']['/resource_a'] = {
            'get': {
                'description': 'resource a',
                'responses': {
                    '200': {'description': 'ok'},
                }
            },
            'post': {
                'description': 'resource a',
                'responses': {
                    '201': {'description': 'ok'},
                }
            },
            'put': {
                'description': 'resource a',
                'responses': {
                    '404': {'description': 'error'},
                }
            }
        }

        renderer = renderers.HttpdomainOldRenderer(
            None,
            {
                'methods': ['post'],
                'paths': ['/resource_a'],
            },
        )
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))

        assert text == textwrap.dedent('''
            .. http:post:: /resource_a
               :synopsis: null

               resource a

               :status 201:
                  ok
        ''').lstrip()

    def test_simpletype_example(self):
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths:
          /resources:
            post:
              summary: Summary
              description: test service
              requestBody:
                description: present
                content:
                  application/json:
                    schema:
                      $ref: '#/components/schemas/MyRequest'
              responses:
                200:
                  description: Success
                  content:
                    application/json:
                      schema:
                        $ref: '#/components/schemas/MyRequest'
        components:
          schemas:
            MyRequest:
              description: absent
              type: object
              properties:
                a:
                  $ref: '#/components/schemas/MySimpleType'
                b:
                  type: string
                  example: B
            MySimpleType:
              type: string
              enum: [A, B, C]
              example: C
        """))

        renderer = renderers.HttpdomainOldRenderer(
            None,
            {
                'examples': True,
                'group_examples': True,
                'entities': True
            }
        )
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent('''
            .. http:post:: /resources
               :synopsis: Summary

               **Summary**

               test service

               :form body: present.
                  Object of type :ref:`MyRequest </components/schemas/MyRequest>`.
               :status 200:
                  Success.
                  Object of type :ref:`MyRequest </components/schemas/MyRequest>`.

               **Example request:**

               .. sourcecode:: http

                  POST /resources HTTP/1.1
                  Host: example.com
                  Content-Type: application/json

                  {
                      "a": "C",
                      "b": "B"
                  }


               **Example response:**

               .. sourcecode:: http

                  HTTP/1.1 200 OK
                  Content-Type: application/json

                  {
                      "a": "C",
                      "b": "B"
                  }

        ''').lstrip()

    def test_header_example(self):
        renderer = renderers.HttpdomainOldRenderer(None, {'examples': True})
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths:
          /resources:
            post:
              summary: Summary
              description: test service
              parameters:
                - name: HeaderRequest
                  in: header
                  description: request header
                  schema:
                    type: string
                    maxLength: 32
                  example: "MD5=thvDyvhfIqlvFe+A9MYgxAfm1q5="
              requestBody:
                content:
                  application/json:
                    schema:
                      type: string
                    example: REQUEST
                    required: true
              responses:
                200:
                  description: Success
                  content:
                    application/json:
                      schema:
                        type: string
                      example: RESPONSE
                  headers:
                    HeaderResponse:
                      description: response header
                      schema:
                        type: string
                      example: "MD5=thvDyvhfIqlvFe+A9MYgxAfm1q5="
        """))

        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent('''
            .. http:post:: /resources
               :synopsis: Summary

               **Summary**

               test service

               :reqheader HeaderRequest:
                  request header.
                  Constraints: maxLength is 32.

               **Example request:**

               .. sourcecode:: http

                  POST /resources HTTP/1.1
                  Host: example.com
                  Content-Type: application/json
                  HeaderRequest: MD5=thvDyvhfIqlvFe+A9MYgxAfm1q5=

                  REQUEST

               :resheader HeaderResponse:
                  response header.
               :status 200:
                  Success.

                  **Example response:**

                  .. sourcecode:: http

                     HTTP/1.1 200 OK
                     Content-Type: application/json
                     HeaderResponse: MD5=thvDyvhfIqlvFe+A9MYgxAfm1q5=

                     RESPONSE

        ''').lstrip()

        renderer = renderers.HttpdomainOldRenderer(
            None,
            {
                'examples': True,
                'group_examples': True
            }
        )
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent('''
            .. http:post:: /resources
               :synopsis: Summary

               **Summary**

               test service

               :reqheader HeaderRequest:
                  request header.
                  Constraints: maxLength is 32.
               :resheader HeaderResponse:
                  response header.
               :status 200:
                  Success.

               **Example request:**

               .. sourcecode:: http

                  POST /resources HTTP/1.1
                  Host: example.com
                  Content-Type: application/json
                  HeaderRequest: MD5=thvDyvhfIqlvFe+A9MYgxAfm1q5=

                  REQUEST


               **Example response:**

               .. sourcecode:: http

                  HTTP/1.1 200 OK
                  Content-Type: application/json
                  HeaderResponse: MD5=thvDyvhfIqlvFe+A9MYgxAfm1q5=

                  RESPONSE

        ''').lstrip()

    def test_entities(self):
        renderer = renderers.HttpdomainOldRenderer(None, {'entities': True})
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths:
          /resources:
            post:
              summary: Summary
              description: test service
              requestBody:
                content:
                  application/json:
                    schema:
                      type: string
                    example: REQUEST
                    required: true
              responses:
                200:
                  description: Success
                  content:
                    application/json:
                      schema:
                        type: string
                      example: RESPONSE
          /resources2:
            post:
              summary: Summary
              description: test service
              requestBody:
                description: Body
                content:
                  application/json:
                    schema:
                      $ref: '#/components/schemas/Test1'
              responses:
                200:
                  description: Success
                  content:
                    application/json:
                      schema:
                        $ref: '#/components/schemas/Test1'
          /resources3:
            post:
              summary: Summary
              description: test service
              requestBody:
                content:
                  application/json:
                    schema:
                      type: array
                      items:
                        $ref: '#/components/schemas/Test1'
              responses:
                200:
                  description: Success
                  content:
                    application/json:
                      schema:
                        type: array
                        items:
                          $ref: '#/components/schemas/Test1'
        components:
          schemas:
            Test1:
              type: object
              properties:
                field1:
                  type: integer
                  format: int32
                  description: Signed 32 bits
        """))

        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent('''
          .. http:post:: /resources
             :synopsis: Summary

             **Summary**

             test service

             :status 200:
                Success.
                Object of type string.

          .. http:post:: /resources2
             :synopsis: Summary

             **Summary**

             test service

             :form body: Body.
                Object of type :ref:`Test1 </components/schemas/Test1>`.
             :status 200:
                Success.
                Object of type :ref:`Test1 </components/schemas/Test1>`.

          .. http:post:: /resources3
             :synopsis: Summary

             **Summary**

             test service

             :form body: Array of :ref:`Test1 </components/schemas/Test1>`.
             :status 200:
                Success.
                Array of :ref:`Test1 </components/schemas/Test1>`.
        ''').lstrip()

        renderer = renderers.HttpdomainOldRenderer(None, {})
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent('''
          .. http:post:: /resources
             :synopsis: Summary

             **Summary**

             test service

             :status 200:
                Success.

          .. http:post:: /resources2
             :synopsis: Summary

             **Summary**

             test service

             :form body: Body.
             :status 200:
                Success.

          .. http:post:: /resources3
             :synopsis: Summary

             **Summary**

             test service

             :status 200:
                Success.
        ''').lstrip()

    def test_adv_parameters(self):
        renderer = renderers.HttpdomainOldRenderer(None, {'entities': True})
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths:
          /resources/{param0}:
            get:
              summary: Summary
              description: test service
              parameters:
                - name: param0
                  in: path
                  description: query param
                  schema:
                    type: string
                    maxLength: 10
                - name: param1
                  in: query
                  description: first param
                  required: true
                  schema:
                    type: string
                    maxLength: 64
                - name: param2
                  in: header
                  description: second param
                  schema:
                    type: string
                    maxLength: 32
                - name: param3
                  in: query
                  description: third param
                  schema:
                    type: array
                    items:
                      type: string
                      maxLength: 64
                    minItems: 1
              responses:
                200:
                  description: Success
                  content:
                    application/json:
                      schema:
                        type: string
                      example: RESPONSE
        """))

        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent('''
          .. http:get:: /resources/{param0}
             :synopsis: Summary

             **Summary**

             test service

             :param string param0:
                query param.
                Object of type string.
                Constraints: maxLength is 10.
             :query string param1:
                first param.
                Object of type string.
                Constraints: maxLength is 64.
                (Required)
             :query array param3:
                third param.
                Array of string.
                Constraints: maxLength is 64; minItems is 1.
             :reqheader param2:
                second param.
                Object of type string.
                Constraints: maxLength is 32.
             :status 200:
                Success.
                Object of type string.
        ''').lstrip()


class TestResolveRefs(object):

    def test_ref_resolving(self):
        data = {
            'foo': {
                'a': 13,
                'b': {
                    'c': True,
                }
            },
            'bar': {
                '$ref': '#/foo/b'
            },
            'baz': [
                {'$ref': '#/foo/a'},
                {'$ref': '#/foo/b'},
                'batman',
            ]
        }

        assert utils._resolve_refs('', data) == {
            'foo': {
                'a': 13,
                'b': {
                    '$entity_ref': '#/foo/b',
                    'c': True,
                }
            },
            'bar': {
                '$entity_ref': '#/foo/b',
                'c': True,
            },
            'baz': [
                13,
                {
                    '$entity_ref': '#/foo/b',
                    'c': True
                },
                'batman',
            ]
        }

    def test_relative_ref_resolving_on_fs(self):
        baseuri = 'file://%s' % os.path.abspath(__file__)

        data = {
            'bar': {
                '$ref': 'testdata/foo.json#/foo/b',
            },
            # check also JSON to YAML references:
            'baz': {
                '$ref': 'testdata/foo.yaml#/foo',
            }
        }

        # import pdb
        # pdb.set_trace()
        assert utils._resolve_refs(baseuri, data) == {
            'bar': {
                '$entity_ref': 'testdata/foo.json#/foo/b',
                'c': True,
            },
            'baz': {
                '$entity_ref': 'testdata/foo.yaml#/foo',
                'a': 17,
                'b': 13,
            },
        }

    def test_noproperties(self):
        renderer = renderers.HttpdomainOldRenderer(None, {'examples': True})
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'openapi': '3.0.0',
            'paths': {
                '/resources': {
                    'post': {
                        'summary': 'Create Resources',
                        'description': '~ some useful description ~',
                        'requestBody': {
                            'content': {
                                'application/json':  {
                                    'schema': {
                                        '$ref': '#/components/schemas/Resource',  # noqa
                                    }
                                }
                            }
                        },
                        'responses': {
                            '200': {
                                'description': 'Something',
                            },
                        },
                    },
                },
            },
            'components': {
                'schemas': {
                    'Resource': {
                        'type': 'object',
                        'additionalProperties': True,
                    },
                },
            },

        }))
        assert text == textwrap.dedent('''
            .. http:post:: /resources
               :synopsis: Create Resources

               **Create Resources**

               ~ some useful description ~


               **Example request:**

               .. sourcecode:: http

                  POST /resources HTTP/1.1
                  Host: example.com
                  Content-Type: application/json

                  {
                      "...": "..."
                  }

               :status 200:
                  Something.
        ''').lstrip()


def test_openapi2_examples(tmpdir, run_sphinx):
    spec = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        'OpenAPI-Specification',
        'examples',
        'v2.0',
        'json',
        'uber.json')
    py.path.local(spec).copy(tmpdir.join('src', 'test-spec.yml'))

    with pytest.raises(ValueError) as excinfo:
        run_sphinx('test-spec.yml', options={'examples': True})

    assert str(excinfo.value) == (
        'Rendering examples is not supported for OpenAPI v2.x specs.')


@pytest.mark.parametrize('render_examples', [False, True])
def test_openapi3_examples(tmpdir, run_sphinx, render_examples):
    spec = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        'OpenAPI-Specification',
        'examples',
        'v3.0',
        'petstore.yaml')
    py.path.local(spec).copy(tmpdir.join('src', 'test-spec.yml'))
    run_sphinx('test-spec.yml', options={'examples': render_examples})

    rendered_html = tmpdir.join('out', 'index.html').read_text('utf-8')

    assert ('<strong>Example response:</strong>' in rendered_html) \
        == render_examples


class TestConvertJsonSchema(object):
    schema = {
        'type': 'object',
        'required': ['name', 'surprise'],
        'properties': {
            'name': {
                'type': 'string',
                'description': 'The name of user'},
            'alias': {
                'type': 'array',
                'items': {
                    'type': 'string'},
                'description': 'The list of user alias'},
            'id': {
                'type': 'integer',
                'description': 'the id of user',
                'readOnly': True},
            'surprise': {
                'type': 'string'},
            'secret': {
                'type': 'string',
                'readOnly': True}}}

    result = list(openapi20.convert_json_schema(schema))

    def test_required_field_with_description(self):
        assert ':<json string name: The name of user (required)' in self.result

    def test_required_field_without_description(self):
        assert ':<json string surprise: (required)' in self.result

    def test_array_field(self):
        assert ':<json string alias[]:' in self.result

    def test_read_only_field_with_description(self):
        assert ':<json integer id: the id of user (read only)' in self.result

    def test_read_only_field_without_description(self):
        assert ':<json string secret: (read only)' in self.result

    def test_nested_schema(self):
        schema = {
            'type': 'object',
            'required': ['name'],
            'properties': {
                'name': {
                    'type': 'string',
                    'description': 'The name of user'
                },
                'friends': {
                    'type': 'array',
                    'items': {
                        'type': 'object',
                        'properties': {
                            'name': {
                                'type': 'string',
                                'readOnly': True
                            },
                            'age': {'type': 'integer'}
                        }
                    },
                    'description': 'The list of user alias'
                },
                'id': {
                    'type': 'integer',
                    'description': 'the id of user',
                    'readOnly': True
                },
                'car': {
                    'type': 'object',
                    'properties': {
                        'provider': {'type': 'string'},
                        'date': {
                            'type': 'string',
                            'description': 'The car of user'
                        }
                    }
                },
                'meta': {
                    'type': 'object',
                    'description': 'free form metadata',
                },
            }
        }

        result = '\n'.join(openapi20.convert_json_schema(schema))

        expected = textwrap.dedent('''
            :<json string car.date: The car of user
            :<json string car.provider:
            :<json integer friends[].age:
            :<json string friends[].name: (read only)
            :<json integer id: the id of user (read only)
            :<json object meta: free form metadata
            :<json string name: The name of user (required)'''.strip('\n'))

        assert result == expected



"""OpenAPI spec renderer: render_request_body."""

import textwrap

import pytest

from sphinxcontrib.openapi import renderers


def textify(generator):
    return "\n".join(generator)


@pytest.mark.parametrize(
    "content_type", ["application/json", "application/foobar+json"]
)
def test_render_request_body_schema_description(
    testrenderer, oas_fragment, content_type
):
    """JSON schema description is rendered."""

    markup = textify(
        testrenderer.render_request_body(
            oas_fragment(
                f"""
                content:
                  {content_type}:
                    schema:
                      properties:
                        foo:
                          type: string
                        bar:
                          type: integer
                """
            ),
            "/evidences/{evidenceId}",
            "POST",
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqjson foo:
        :reqjsonobj foo: string
        :reqjson bar:
        :reqjsonobj bar: integer

        """
    )


def test_render_request_body_schema_description_non_json(testrenderer, oas_fragment):
    """JSON schema is not rendered for non JSON mimetype."""

    markup = textify(
        testrenderer.render_request_body(
            oas_fragment(
                """
                content:
                  text/csv:
                    schema:
                      properties:
                        foo:
                          type: string
                        bar:
                          type: integer
                """
            ),
            "/evidences/{evidenceId}",
            "POST",
        )
    )
    assert markup == textwrap.dedent(
        """\
        """
    )


def test_render_request_body_schema_description_turned_off(fakestate, oas_fragment):
    """JSON schema description is not rendered b/c feature is off."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"no-json-schema-description": True},
    )

    markup = textify(
        testrenderer.render_request_body(
            oas_fragment(
                """
                content:
                  application/json:
                    schema:
                      properties:
                        foo:
                          type: string
                        bar:
                          type: integer
                """
            ),
            "/evidences/{evidenceId}",
            "POST",
        )
    )
    assert markup == textwrap.dedent(
        """\
        """
    )



"""Partial OpenAPI v2.x (fka Swagger) to OpenAPI v3.x converter."""

import functools
import urllib

import picobox


__all__ = [
    "convert",
]


def convert(spec):
    """Convert a given OAS 2 spec to OAS 3."""

    return Lib2to3().convert(spec)


def _is_vendor_extension(key):
    """Return 'True' if a given key is a vendor extension."""

    return key.startswith("x-")


def _get_properties(node, properties, *, vendor_extensions=False):
    """Return a subset of 'node' properties w/ or wo/ vendor extensions."""

    return {
        key: value
        for key, value in node.items()
        if any([key in properties, vendor_extensions and _is_vendor_extension(key)])
    }


def _get_schema_properties(node, *, except_for=None):
    """Find and return 'Schema Object' properties."""

    except_for = except_for or set()
    schema = _get_properties(
        node,
        {
            "additionalProperties",
            "allOf",
            "default",
            "description",
            "discriminator",
            "enum",
            "example",
            "exclusiveMaximum",
            "exclusiveMinimum",
            "externalDocs",
            "format",
            "items",
            "maxItems",
            "maxLength",
            "maxProperties",
            "maximum",
            "minItems",
            "minLength",
            "minProperties",
            "minimum",
            "multipleOf",
            "pattern",
            "properties",
            "readOnly",
            "required",
            "title",
            "type",
            "uniqueItems",
            "xml",
        }
        - set(except_for),
    )

    if "discriminator" in schema:
        schema["discriminator"] = {"propertyName": schema["discriminator"]}
    return schema


def _items_wo_vendor_extensions(node):
    """Iterate over 'node' properties excluding vendor extensions."""

    for key, value in node.items():
        if _is_vendor_extension(key):
            continue
        yield key, value


class Lib2to3:

    _target_version = "3.0.3"
    _injector = picobox.Stack()

    def _insert_into_injector(name):
        def decorator(fn):
            @functools.wraps(fn)
            def wrapper(self, node, *args, **kwargs):
                with Lib2to3._injector.push(picobox.Box(), chain=True) as box:
                    box.put(name, factory=lambda: node, scope=picobox.threadlocal)
                    return fn(self, node, *args, **kwargs)

            return wrapper

        return decorator

    def __init__(self):
        self._schemes = set()

    @_insert_into_injector("spec")
    def convert(self, spec):
        # The following OAS 2 fields are ignored and not converted. Mostly due
        # to the fact that we expect *resolved* spec as input, and most of its
        # fields are used to group shared (i.e. referenced) objects that will
        # not exist in the resolved spec.
        #
        #  - definitions
        #  - parameters
        #  - responses
        #  - securityDefinitions
        #  - security
        #
        # By no means one must assume that these fields will never be
        # converted. I simply have no time to work on this, and for
        # sphixcontrib-openapi purposes it's not actually needed.

        converted = {
            "info": spec["info"],
            "openapi": self._target_version,
            "paths": self.convert_paths(spec["paths"]),
        }
        converted.update(
            _get_properties(spec, {"tags", "externalDocs"}, vendor_extensions=True),
        )

        servers = self.convert_servers(spec)
        if servers:
            converted["servers"] = servers

        return converted

    @_insert_into_injector("paths")
    def convert_paths(self, paths):
        converted = _get_properties(paths, {}, vendor_extensions=True)

        for endpoint, path in _items_wo_vendor_extensions(paths):
            converted[endpoint] = self.convert_path(path)

        return converted

    @_insert_into_injector("path")
    def convert_path(self, path):
        converted = _get_properties(path, {}, vendor_extensions=True)

        for key, value in _items_wo_vendor_extensions(path):
            if key == "parameters":
                converted[key] = self.convert_parameters(value)
            else:
                converted[key] = self.convert_operation(value)

        return converted

    @_insert_into_injector("operation")
    def convert_operation(self, operation):
        converted = _get_properties(
            operation,
            {
                "tags",
                "summary",
                "description",
                "externalDocs",
                "operationId",
                "deprecated",
                "security",
            },
            vendor_extensions=True,
        )

        # Memorize every encountered 'schemes'. Since this property does not
        # exist in OAS 3, it seems the best we can do is to use them in OAS 3
        # 'servers' object.
        self._schemes.update(operation.get("schemes", []))

        if "parameters" in operation:
            parameters = self.convert_parameters(operation["parameters"])

            # Both 'body' and 'formData' parameters are mutually exclusive,
            # therefore there's no way we may end up with both kinds at once.
            request_body = self.convert_request_body(operation)
            request_body = request_body or self.convert_request_body_formdata(operation)

            if parameters:
                converted["parameters"] = parameters

            if request_body:
                converted["requestBody"] = request_body

        converted["responses"] = self.convert_responses(operation["responses"])
        return converted

    @_injector.pass_("spec")
    def convert_request_body(self, operation, *, spec):
        # OAS 3 expects an explicitly specified mimetype of the request body.
        # It's not clear what to do if OAS 2 'consumes' is not defined. Let's
        # start with a glob pattern and figure out what a better option could
        # be later on.
        consumes = operation.get("consumes") or spec.get("consumes") or ["*/*"]

        for parameter in operation["parameters"]:
            if parameter["in"] == "body":
                # Since 'requestBody' is completely new and nested object in
                # OAS 3, it's not clear what should we insert possible vendor
                # extensions. Thus, let's ignore them until we figure it out.
                converted = _get_properties(parameter, {"description", "required"})
                converted["content"] = {
                    consume: {"schema": parameter["schema"]} for consume in consumes
                }
                return converted

        return None

    @_injector.pass_("spec")
    def convert_request_body_formdata(self, operation, *, spec):
        consumes = (
            operation.get("consumes")
            or spec.get("consumes")
            or ["application/x-www-form-urlencoded"]
        )
        supported = {
            "application/x-www-form-urlencoded",
            "multipart/form-data",
        }
        mimetypes = supported.intersection(consumes)
        schema = {"type": "object", "properties": {}}

        for parameter in operation["parameters"]:
            if parameter["in"] == "formData":
                schema["properties"][parameter["name"]] = _get_schema_properties(
                    parameter, except_for={"name", "in", "required"}
                )

                if parameter.get("required"):
                    schema.setdefault("required", []).append(parameter["name"])

                # Excerpt from OpenAPI 2.x spec:
                #
                # > If type is "file", the consumes MUST be either
                # > "multipart/form-data", "application/x-www-form-urlencoded"
                # > or both and the parameter MUST be in "formData".
                #
                # This is weird since HTTP does not allow file uploading in
                # 'application/x-www-form-urlencoded'. Moreover, Swagger
                # editor complains if 'file' is detected and there's no
                # 'multipart/form-data' in `consumes'.
                if parameter["type"] == "file":
                    mimetypes = ["multipart/form-data"]

        if not schema["properties"]:
            return None
        return {"content": {mimetype: {"schema": schema} for mimetype in mimetypes}}

    @_insert_into_injector("parameters")
    def convert_parameters(self, parameters):
        return [
            self.convert_parameter(parameter)
            for parameter in parameters
            # If a parameter is one of the backward compatible type, delegate
            # the call to the converter function. Incompatible types, such as
            # 'formData' and 'body', must be handled separately since they are
            # reflected in 'Operation Object' in OAS 3.
            if parameter["in"] in {"query", "header", "path"}
        ]

    @_insert_into_injector("parameter")
    def convert_parameter(self, parameter):
        schema = _get_schema_properties(
            parameter,
            # Some of 'Parameter Object' properties have the same name as some
            # of 'Schema Object' properties. Since we know for sure that in
            # this context they are part of 'Parameter Object', we should
            # ignore their meaning as part of 'Schema Object'.
            except_for={"name", "in", "description", "required"},
        )
        converted = {
            key: value for key, value in parameter.items() if key not in schema
        }
        converted["schema"] = schema
        collection_format = converted.pop("collectionFormat", None)

        if converted["in"] in {"path", "header"} and collection_format == "csv":
            converted["style"] = "simple"
        elif converted["in"] in {"query"} and collection_format:
            styles = {
                "csv": {"style": "form", "explode": False},
                "multi": {"style": "form", "explode": True},
                "ssv": {"style": "spaceDelimited"},
                "pipes": {"style": "pipeDelimited"},
                # OAS 3 does not explicitly say what is the alternative to
                # 'collectionFormat=tsv'. We have no other option but to ignore
                # it. Fortunately, we don't care much as it's not used by the
                # renderer.
                "tsv": {},
            }
            converted.update(styles[collection_format])

        return converted

    @_insert_into_injector("responses")
    def convert_responses(self, responses):
        converted = _get_properties(responses, {}, vendor_extensions=True)

        for status_code, response in _items_wo_vendor_extensions(responses):
            converted[status_code] = self.convert_response(response)

        return converted

    @_injector.pass_("spec")
    @_injector.pass_("operation")
    @_insert_into_injector("response")
    def convert_response(self, response, *, spec, operation):
        converted = _get_properties(response, {"description"}, vendor_extensions=True)

        # OAS 3 expects an explicitly specified mimetype in the response. It's
        # not clear what to do if OAS 2 'produces' is not defined. Let's start
        # with a glob pattern and figure out what a better option could be
        # later on.
        produces = operation.get("produces") or spec.get("produces") or ["*/*"]
        schema = response.get("schema")
        examples = response.get("examples")

        if schema or examples:
            content = converted.setdefault("content", {})

            if schema is not None:
                for mimetype in produces:
                    content.setdefault(mimetype, {})["schema"] = schema

            if examples is not None:
                # According to OAS2, mimetypes in 'examples' property MUST be
                # one of the operation's 'produces'.
                for mimetype, example in examples.items():
                    content.setdefault(mimetype, {})["example"] = example

        if "headers" in response:
            converted["headers"] = {
                key: dict(
                    _get_properties(value, "description", vendor_extensions=True),
                    schema=_get_schema_properties(value, except_for={"description"}),
                )
                for key, value in response["headers"].items()
            }

        return converted

    def convert_servers(self, spec):
        """Convert OAS2 '(host, basePath, schemes)' triplet into OAS3 'servers' node."""

        host = spec.get("host", "")
        basepath = spec.get("basePath", "")
        schemes = self._schemes.union(spec.get("schemes", set()))

        # Since 'host', 'basePath' and 'schemes' are optional in OAS 2, there
        # may be the case when they aren't set. If that's happened it means
        # there's nothing to convert, and thus we simply return an empty list.
        if not host and not basepath and not schemes:
            return []

        if not schemes:
            # If 'host' is not set, the url will contain a bare basePath.
            # According to OAS 3, it's a valid URL, and both the host and the
            # scheme must be assumed to be the same as the server that shared
            # this OAS 3 spec.
            return [{"url": urllib.parse.urljoin(host, basepath)}]

        return [
            {"url": urllib.parse.urlunsplit([scheme, host, basepath, None, None])}
            for scheme in sorted(schemes)
        ]



""".convert_path() test suite."""

import pytest

import sphinxcontrib.openapi._lib2to3 as lib2to3


@pytest.fixture(scope="function")
def convert_path(oas_fragment):
    def _wrapper(path):
        oas2 = oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: "1.0"
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            """
        )
        oas2["paths"]["/test"] = path

        oas3 = lib2to3.convert(oas2)
        return oas3["paths"]["/test"]

    return _wrapper


@pytest.mark.parametrize(
    "method", ["get", "put", "post", "delete", "options", "head", "patch"]
)
def test_minimal(convert_path, oas_fragment, method):
    converted = convert_path(
        oas_fragment(
            f"""
            {method}:
              responses:
                '200':
                  description: a response description
            """
        ),
    )
    assert converted == oas_fragment(
        f"""
        {method}:
          responses:
            '200':
              description: a response description
        """
    )


def test_complete(convert_path, oas_fragment):
    converted = convert_path(
        oas_fragment(
            """
            parameters:
              - in: path
                name: username
                required: true
                type: string
            get:
              tags:
                - tag_a
                - tag_b
              summary: an operation summary
              description: an operation description
              externalDocs: https://docs.example.com/
              operationId: myOperation
              produces:
                - application/json
              parameters:
                - in: header
                  name: token
                  type: string
                - in: query
                  name: id
                  type: string
              responses:
                '200':
                  schema:
                    items:
                      format: int32
                      type: integer
                    type: array
                  description: a response description
            """
        ),
    )
    assert converted == oas_fragment(
        """
        parameters:
          - in: path
            name: username
            schema:
              type: string
            required: true
        get:
          tags:
            - tag_a
            - tag_b
          summary: an operation summary
          description: an operation description
          externalDocs: https://docs.example.com/
          operationId: myOperation
          parameters:
            - in: header
              name: token
              schema:
                type: string
            - in: query
              name: id
              schema:
                type: string
          responses:
            '200':
              content:
                application/json:
                  schema:
                    items:
                      format: int32
                      type: integer
                    type: array
              description: a response description
        """
    )


def test_shared_parameters(convert_path, oas_fragment):
    converted = convert_path(
        oas_fragment(
            """
            parameters:
              - in: path
                name: username
                required: true
                type: string
            get:
              parameters:
                - in: header
                  name: token
                  type: string
                - in: query
                  name: id
                  type: string
              responses:
                '200':
                  description: a response description
            """
        ),
    )
    assert converted == oas_fragment(
        """
        parameters:
          - in: path
            name: username
            schema:
              type: string
            required: true
        get:
          parameters:
            - in: header
              name: token
              schema:
                type: string
            - in: query
              name: id
              schema:
                type: string
          responses:
            '200':
              description: a response description
        """
    )


def test_multiple(convert_path, oas_fragment):
    converted = convert_path(
        oas_fragment(
            """
            post:
              responses:
                '201':
                  description: a post response description
            get:
              responses:
                '200':
                  description: a get response description
            """
        ),
    )
    assert converted == oas_fragment(
        """
        post:
          responses:
            '201':
              description: a post response description
        get:
          responses:
            '200':
              description: a get response description
        """
    )


def test_vendor_extensions(convert_path, oas_fragment):
    converted = convert_path(
        oas_fragment(
            """
            get:
              responses:
                '200':
                  description: a response description
            x-vendor-ext: vendor-ext
            """
        ),
    )
    assert converted == oas_fragment(
        """
        get:
          responses:
            '200':
              description: a response description
        x-vendor-ext: vendor-ext
        """
    )



"""OpenAPI spec renderer: render_responses."""

import textwrap

from sphinxcontrib.openapi import renderers


def textify(generator):
    return "\n".join(generator)


def test_render_responses_no_items(testrenderer, oas_fragment):
    """No response definitions are rendered."""

    markup = textify(
        testrenderer.render_responses(
            oas_fragment(
                """
                {}
                """
            )
        )
    )
    assert markup == ""


def test_render_responses_one_item(testrenderer, oas_fragment):
    """One usual response definition is rendered."""

    markup = textify(
        testrenderer.render_responses(
            oas_fragment(
                """
                '200':
                  description: An evidence.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.
        """.rstrip()
    )


def test_render_responses_one_item_status_code_int(testrenderer, oas_fragment):
    """One usual response definition is rendered even if status code is integer."""

    markup = textify(
        testrenderer.render_responses(
            oas_fragment(
                """
                200:
                  description: An evidence.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.
        """.rstrip()
    )


def test_render_responses_many_items(testrenderer, oas_fragment):
    """Many response definitions are rendered."""

    markup = textify(
        testrenderer.render_responses(
            oas_fragment(
                """
                '200':
                  description: An evidence.
                '404':
                  description: An evidence not found.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.
        :statuscode 404:
           An evidence not found.
        """.rstrip()
    )


def test_render_responses_json_schema_description(testrenderer, oas_fragment):
    """JSON schema description is rendered."""

    markup = textify(
        testrenderer.render_responses(
            oas_fragment(
                """
                '200':
                  description: An evidence.
                  content:
                    application/json:
                      schema:
                        properties:
                          foo:
                            type: string
                          bar:
                            type: integer
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :resjson foo:
        :resjsonobj foo: string
        :resjson bar:
        :resjsonobj bar: integer

        :statuscode 200:
           An evidence.
        """
    )


def test_render_responses_json_schema_description_4xx(testrenderer, oas_fragment):
    """JSON schema description is rendered."""

    markup = textify(
        testrenderer.render_responses(
            oas_fragment(
                """
                '400':
                  description: An evidence.
                  content:
                    application/json:
                      schema:
                        properties:
                          foo:
                            type: string
                          bar:
                            type: integer
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 400:
           An evidence.
        """.rstrip()
    )


def test_render_responses_json_schema_description_first_2xx(testrenderer, oas_fragment):
    """JSON schema description is rendered."""

    markup = textify(
        testrenderer.render_responses(
            oas_fragment(
                """
                '400':
                  description: An error.
                  content:
                    application/json:
                      schema:
                        properties:
                          aaa:
                            type: string
                '200':
                  description: An evidence.
                  content:
                    application/json:
                      schema:
                        properties:
                          foo:
                            type: string
                          bar:
                            type: integer
                '201':
                  description: An evidence created.
                  content:
                    application/json:
                      schema:
                        properties:
                          bbb:
                            type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :resjson foo:
        :resjsonobj foo: string
        :resjson bar:
        :resjsonobj bar: integer

        :statuscode 400:
           An error.
        :statuscode 200:
           An evidence.

        :statuscode 201:
           An evidence created.
        """
    )


def test_render_responses_json_schema_description_turned_off(fakestate, oas_fragment):
    """JSON schema description is not rendered b/c feature is off."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"no-json-schema-description": True},
    )

    markup = textify(
        testrenderer.render_responses(
            oas_fragment(
                """
                '200':
                  description: An evidence.
                  content:
                    application/json:
                      schema:
                        properties:
                          foo:
                            type: string
                          bar:
                            type: integer
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.
        """
    )



import os

# Unfortunately, Sphinx doesn't support code highlighting for standard
# reStructuredText `code` directive. So let's register 'code' directive
# as alias for Sphinx's own implementation.
#
# https://github.com/sphinx-doc/sphinx/issues/2155
from docutils.parsers.rst import directives
from sphinx.directives.code import CodeBlock

directives.register_directive("code", CodeBlock)

project = "sphinxcontrib-openapi"
copyright = "2016, Ihor Kalnytskyi"

extensions = ["sphinxcontrib.openapi"]
source_suffix = ".rst"
master_doc = "index"
exclude_patterns = ["_build"]
pygments_style = "default"

if not os.environ.get("READTHEDOCS") == "True":
    import sphinx_rtd_theme

    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]



import textwrap
import collections

from sphinxcontrib.openapi import renderers


class TestOpenApi3HttpDomain(object):

    def test_basic(self):
        renderer = renderers.TocRenderer(None, {})
        spec = collections.defaultdict(collections.OrderedDict)
        spec['paths']['/resource_a'] = {
            'get': {
                'description': 'resource a',
                'responses': {
                    '200': {'description': 'ok'},
                }
            }
        }
        spec['paths']['/resource_b'] = {
            'post': {
                'operationId': 'UpdateResourceB',
                'description': 'resource b',
                'responses': {
                    '404': {'description': 'error'},
                }
            }
        }

        text = '\n'.join(renderer.render_restructuredtext_markup(spec))

        assert text == textwrap.dedent("""
        .. hlist::
            :columns: 2

            - `get /resource_a <#get--resource_a>`_
            - `UpdateResourceB <#post--resource_b>`_
        """)

    def test_options(self):
        renderer = renderers.TocRenderer(None, {"nb_columns": 3})
        spec = collections.defaultdict(collections.OrderedDict)
        spec['paths']['/resource_a'] = {
            'get': {
                'description': 'resource a',
                'responses': {
                    '200': {'description': 'ok'},
                }
            }
        }
        spec['paths']['/resource_b'] = {
            'post': {
                'operationId': 'UpdateResourceB',
                'description': 'resource b',
                'responses': {
                    '404': {'description': 'error'},
                }
            }
        }

        text = '\n'.join(renderer.render_restructuredtext_markup(spec))

        assert text == textwrap.dedent("""
        .. hlist::
            :columns: 3

            - `get /resource_a <#get--resource_a>`_
            - `UpdateResourceB <#post--resource_b>`_
        """)



""".convert_request_body() test suite."""

import pytest

import sphinxcontrib.openapi._lib2to3 as lib2to3


@pytest.fixture(scope="function")
def convert_request_body(oas_fragment):
    def _wrapper(operation_fragment):
        oas2 = oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: "1.0"
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            """
        )
        oas2["paths"]["/test"]["get"].update(operation_fragment)

        oas3 = lib2to3.convert(oas2)
        return oas3["paths"]["/test"]["get"]["requestBody"]

    return _wrapper


def test_minimal(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            parameters:
              - in: formData
                name: user
                type: string
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          application/x-www-form-urlencoded:
            schema:
              properties:
                user:
                  type: string
              type: object
        """
    )


def test_complete(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            parameters:
              - description: a name of the user
                in: formData
                name: user
                type: string
                required: true
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          application/x-www-form-urlencoded:
            schema:
              type: object
              properties:
                user:
                  description: a name of the user
                  type: string
              required: [user]
        """
    )


def test_complex_schema(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            parameters:
              - format: int32
                in: formData
                maximum: 100
                minimum: 5
                name: age
                type: integer
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          application/x-www-form-urlencoded:
            schema:
              properties:
                age:
                  format: int32
                  maximum: 100
                  minimum: 5
                  type: integer
              type: object
        """
    )


def test_consumes_urlencoded(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            consumes:
              - application/x-www-form-urlencoded
            parameters:
              - description: a name of the user
                in: formData
                name: user
                type: string
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          application/x-www-form-urlencoded:
            schema:
              properties:
                user:
                  description: a name of the user
                  type: string
              type: object
        """
    )


def test_consumes_form_data(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            consumes:
              - multipart/form-data
            parameters:
              - description: a name of the user
                in: formData
                name: user
                type: string
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          multipart/form-data:
            schema:
              properties:
                user:
                  description: a name of the user
                  type: string
              type: object
        """
    )


def test_consumes_urlencoded_and_form_data(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            consumes:
              - application/x-www-form-urlencoded
              - multipart/form-data
            parameters:
              - description: a name of the user
                in: formData
                name: user
                type: string
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          application/x-www-form-urlencoded:
            schema:
              properties:
                user:
                  description: a name of the user
                  type: string
              type: object
          multipart/form-data:
            schema:
              properties:
                user:
                  description: a name of the user
                  type: string
              type: object
        """
    )


def test_required(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            parameters:
              - description: a name of the user
                in: formData
                name: user
                required: true
                type: string
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          application/x-www-form-urlencoded:
            schema:
              properties:
                user:
                  description: a name of the user
                  type: string
              required:
                - user
              type: object
        """
    )


def test_multiple(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            parameters:
              - description: a name of the user
                in: formData
                name: user
                type: string
              - description: a status of the user
                in: formData
                name: status
                type: string
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          application/x-www-form-urlencoded:
            schema:
              properties:
                status:
                  description: a status of the user
                  type: string
                user:
                  description: a name of the user
                  type: string
              type: object
        """
    )


def test_type_file_implicit_form_data(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            parameters:
              - description: a user pic
                in: formData
                name: userpic
                type: file
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          multipart/form-data:
            schema:
              properties:
                userpic:
                  description: a user pic
                  type: file
              type: object
        """
    )


def test_type_file_consumes_form_data(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            consumes:
              - multipart/form-data
            parameters:
              - description: a user pic
                in: formData
                name: userpic
                type: file
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          multipart/form-data:
            schema:
              properties:
                userpic:
                  description: a user pic
                  type: file
              type: object
        """
    )


def test_consumes_json_and_urlencoded(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            consumes:
              - application/json
              - application/x-www-form-urlencoded
            parameters:
              - description: a name of the user
                in: formData
                name: user
                type: string
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          application/x-www-form-urlencoded:
            schema:
              properties:
                user:
                  description: a name of the user
                  type: string
              type: object
        """
    )


def test_consumes_json_and_form_data(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            consumes:
              - application/json
              - multipart/form-data
            parameters:
              - description: a name of the user
                in: formData
                name: user
                type: string
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          multipart/form-data:
            schema:
              properties:
                user:
                  description: a name of the user
                  type: string
              type: object
        """
    )



"""OpenAPI spec renderer: render_json_schema_description."""

import textwrap

import pytest

from sphinxcontrib.openapi import renderers


def textify(generator):
    return "\n".join(generator)


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_root_object(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for JSON object in root."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  prop_a:
                    type: string
                  prop_b:
                    type: object
                    properties:
                      eggs:
                        type: boolean
                  prop_c:
                    type: number
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} prop_a:
        :{typedirective} prop_a: string
        :{directive} prop_b:
        :{typedirective} prop_b: object
        :{directive} prop_b.eggs:
        :{typedirective} prop_b.eggs: boolean
        :{directive} prop_c:
        :{typedirective} prop_c: number
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjsonarr", "reqjsonarrtype", id="req"),
        pytest.param("res", "resjsonarr", "resjsonarrtype", id="res"),
    ],
)
def test_render_json_schema_description_root_array(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for JSON array in root."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: array
                items:
                  type: object
                  properties:
                    prop:
                      type: string
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} prop:
        :{typedirective} prop: string
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
@pytest.mark.parametrize(
    ["schema_type"],
    [
        pytest.param("null"),
        pytest.param("boolean"),
        pytest.param("number"),
        pytest.param("string"),
        pytest.param("integer"),
    ],
)
def test_render_json_schema_description_root_unsupported(
    testrenderer, oas_fragment, schema_type, req_or_res, directive, typedirective
):
    """JSON schema description is not generated for unsupported type in root."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                f"""
                type: {schema_type}
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        """\
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_root_any_of_object(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for anyOf JSON object in root."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                anyOf:
                  - type: object
                    properties:
                      prop_a:
                        type: string
                      prop_b:
                        type: number
                  - type: object
                    properties:
                      prop_c:
                        type: object
                        properties:
                          eggs:
                            type: boolean
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} prop_a:
        :{typedirective} prop_a: string
        :{directive} prop_b:
        :{typedirective} prop_b: number
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjsonarr", "reqjsonarrtype", id="req"),
        pytest.param("res", "resjsonarr", "resjsonarrtype", id="res"),
    ],
)
def test_render_json_schema_description_root_any_of_array(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for anyOf JSON array in root."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                anyOf:
                  - type: array
                    items:
                      type: object
                      properties:
                        prop:
                          type: string
                  - type: array
                    items:
                      type: object
                      properties:
                        prop:
                          type: number
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} prop:
        :{typedirective} prop: string
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
@pytest.mark.parametrize(
    ["schema_type"],
    [
        pytest.param("null"),
        pytest.param("boolean"),
        pytest.param("number"),
        pytest.param("string"),
        pytest.param("integer"),
    ],
)
def test_render_json_schema_description_root_any_of_unsupported(
    testrenderer, oas_fragment, schema_type, req_or_res, directive, typedirective
):
    """JSON schema description is not generated for anyOf unsupported type in root."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                f"""
                anyOf:
                  - type: {schema_type}
                  - type: object
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        """\
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_root_one_of_object(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for oneOf JSON object in root."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                oneOf:
                  - type: object
                    properties:
                      prop_a:
                        type: string
                      prop_b:
                        type: number
                  - type: object
                    properties:
                      prop_c:
                        type: object
                        properties:
                          eggs:
                            type: boolean
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} prop_a:
        :{typedirective} prop_a: string
        :{directive} prop_b:
        :{typedirective} prop_b: number
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjsonarr", "reqjsonarrtype", id="req"),
        pytest.param("res", "resjsonarr", "resjsonarrtype", id="res"),
    ],
)
def test_render_json_schema_description_root_one_of_array(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for oneOf JSON array in root."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                oneOf:
                  - type: array
                    items:
                      type: object
                      properties:
                        prop:
                          type: string
                  - type: array
                    items:
                      type: object
                      properties:
                        prop:
                          type: number
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} prop:
        :{typedirective} prop: string
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
@pytest.mark.parametrize(
    ["schema_type"],
    [
        pytest.param("null"),
        pytest.param("boolean"),
        pytest.param("number"),
        pytest.param("string"),
        pytest.param("integer"),
    ],
)
def test_render_json_schema_description_root_one_of_unsupported(
    testrenderer, oas_fragment, schema_type, req_or_res, directive, typedirective
):
    """JSON schema description is not generated for oneOf unsupported type in root."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                f"""
                oneOf:
                  - type: {schema_type}
                  - type: object
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        """\
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_root_all_of_object(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for allOf in root."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                allOf:
                  - properties:
                      name:
                        properties:
                          first:
                            type: string
                      age:
                        type: integer
                  - properties:
                      name:
                        properties:
                          last:
                            type: string
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} name:
        :{typedirective} name: object
        :{directive} name.first:
        :{typedirective} name.first: string
        :{directive} name.last:
        :{typedirective} name.last: string
        :{directive} age:
        :{typedirective} age: integer
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
@pytest.mark.parametrize(
    ["schema_type"],
    [
        pytest.param("null"),
        pytest.param("boolean"),
        pytest.param("number"),
        pytest.param("string"),
        pytest.param("integer"),
    ],
)
def test_render_json_schema_description_primitive(
    testrenderer, oas_fragment, schema_type, req_or_res, directive, typedirective
):
    """JSON schema description is generated for primitive types."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                f"""
                type: object
                properties:
                  some_key:
                    type: "{schema_type}"
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} some_key:
        :{typedirective} some_key: {schema_type}
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_object(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for JSON object."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  root:
                    type: object
                    properties:
                      prop_a:
                        type: string
                      prop_b:
                        type: object
                        properties:
                          eggs:
                            type: boolean
                      prop_c:
                        type: number
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} root:
        :{typedirective} root: object
        :{directive} root.prop_a:
        :{typedirective} root.prop_a: string
        :{directive} root.prop_b:
        :{typedirective} root.prop_b: object
        :{directive} root.prop_b.eggs:
        :{typedirective} root.prop_b.eggs: boolean
        :{directive} root.prop_c:
        :{typedirective} root.prop_c: number
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_object_implicit(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for implicit JSON object."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  root:
                    properties:
                      prop_a:
                        type: string
                      prop_b:
                        properties:
                          eggs:
                            type: boolean
                      prop_c:
                        type: number
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} root:
        :{typedirective} root: object
        :{directive} root.prop_a:
        :{typedirective} root.prop_a: string
        :{directive} root.prop_b:
        :{typedirective} root.prop_b: object
        :{directive} root.prop_b.eggs:
        :{typedirective} root.prop_b.eggs: boolean
        :{directive} root.prop_c:
        :{typedirective} root.prop_c: number
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_array(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for JSON array."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  root:
                    type: array
                    items:
                      type: object
                      properties:
                        prop_a:
                          type: string
                        prop_b:
                          type: array
                          items:
                            type: number
                        prop_c:
                          type: number
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} root[]:
        :{typedirective} root[]: object
        :{directive} root[].prop_a:
        :{typedirective} root[].prop_a: string
        :{directive} root[].prop_b[]:
        :{typedirective} root[].prop_b[]: number
        :{directive} root[].prop_c:
        :{typedirective} root[].prop_c: number
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_array_implicit(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for implicit JSON array."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  root:
                    items:
                      type: object
                      properties:
                        prop_a:
                          type: string
                        prop_b:
                          items:
                            type: number
                        prop_c:
                          type: number
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} root[]:
        :{typedirective} root[]: object
        :{directive} root[].prop_a:
        :{typedirective} root[].prop_a: string
        :{directive} root[].prop_b[]:
        :{typedirective} root[].prop_b[]: number
        :{directive} root[].prop_c:
        :{typedirective} root[].prop_c: number
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_format(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for formatted types."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  created_at:
                    type: string
                    format: date-time
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} created_at:
        :{typedirective} created_at: string:date-time
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_deprecated(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated with deprecated marker."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  created_at:
                    type: string
                    deprecated: true
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} created_at:
        :{typedirective} created_at: string, deprecated
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_required(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for JSON object w/ required marker."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  root:
                    type: object
                    properties:
                      prop_a:
                        type: string
                      prop_b:
                        type: object
                        properties:
                          eggs:
                            type: boolean
                        required: [eggs]
                      prop_c:
                        type: number
                    required: [prop_a]
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} root:
        :{typedirective} root: object
        :{directive} root.prop_a:
        :{typedirective} root.prop_a: string, required
        :{directive} root.prop_b:
        :{typedirective} root.prop_b: object
        :{directive} root.prop_b.eggs:
        :{typedirective} root.prop_b.eggs: boolean, required
        :{directive} root.prop_c:
        :{typedirective} root.prop_c: number
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_deprecated_and_required(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for JSON object w/ deprecated & required markers."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  root:
                    type: object
                    properties:
                      prop_a:
                        type: string
                      prop_b:
                        type: object
                        properties:
                          eggs:
                            type: boolean
                            deprecated: true
                        required: [eggs]
                      prop_c:
                        type: number
                    required: [prop_a]
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} root:
        :{typedirective} root: object
        :{directive} root.prop_a:
        :{typedirective} root.prop_a: string, required
        :{directive} root.prop_b:
        :{typedirective} root.prop_b: object
        :{directive} root.prop_b.eggs:
        :{typedirective} root.prop_b.eggs: boolean, deprecated, required
        :{directive} root.prop_c:
        :{typedirective} root.prop_c: number
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_description(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated with description."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                description: a resource representation
                properties:
                  created_at:
                    type: string
                    description: a resource creation time
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} created_at:
           a resource creation time
        :{typedirective} created_at: string
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_description_commonmark_default(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated with CommonMark description by default."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                description: a resource representation
                properties:
                  created_at:
                    type: string
                    description: a `resource` creation __time__
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} created_at:
           a ``resource`` creation **time**
        :{typedirective} created_at: string
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_description_commonmark(
    fakestate, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated with CommonMark description."""

    testrenderer = renderers.HttpdomainRenderer(fakestate, {"markup": "commonmark"})
    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                description: a resource representation
                properties:
                  created_at:
                    type: string
                    description: a `resource` creation __time__
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} created_at:
           a ``resource`` creation **time**
        :{typedirective} created_at: string
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_description_restructuredtext(
    fakestate, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated with reStructuredText description."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"markup": "restructuredtext"}
    )
    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                description: a resource representation
                properties:
                  created_at:
                    type: string
                    description: a `resource` creation __time__
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} created_at:
           a `resource` creation __time__
        :{typedirective} created_at: string
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_any_of(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for anyOf."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                f"""
                type: object
                properties:
                  some_key:
                    anyOf:
                      - type: integer
                      - type: string
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} some_key:
        :{typedirective} some_key: integer
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_one_of(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for oneOf."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  some_key:
                    oneOf:
                      - type: integer
                      - type: string
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} some_key:
        :{typedirective} some_key: integer
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_all_of(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for allOf."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  person:
                    allOf:
                      - properties:
                          name:
                            properties:
                              first:
                                type: string
                          age:
                            type: integer
                      - properties:
                          name:
                            properties:
                              last:
                                type: string
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} person:
        :{typedirective} person: object
        :{directive} person.name:
        :{typedirective} person.name: object
        :{directive} person.name.first:
        :{typedirective} person.name.first: string
        :{directive} person.name.last:
        :{typedirective} person.name.last: string
        :{directive} person.age:
        :{typedirective} person.age: integer
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_all_of_logical_impossible(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for allOf that is logical impossible."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  some_key:
                    allOf:
                      - type: integer
                      - type: string
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} some_key:
        :{typedirective} some_key: string
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_any_of_shared_type(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for anyOf w/ shared 'type'."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  some_key:
                    type: string
                    anyOf:
                      - minLength: 3
                      - maxLength: 5
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} some_key:
        :{typedirective} some_key: string
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_one_of_shared_type(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for oneOf w/ shared 'type'."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  some_key:
                    type: string
                    oneOf:
                      - minLength: 3
                      - maxLength: 5
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} some_key:
        :{typedirective} some_key: string
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_all_of_shared_type(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for allOf w/ shared 'type'."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  some_key:
                    type: string
                    alOf:
                      - minLength: 3
                      - maxLength: 5
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} some_key:
        :{typedirective} some_key: string
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_not(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for JSON *not*."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  root:
                    not:
                      type: boolean
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} root:
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_enum(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for JSON enum."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  root:
                    type: string
                    enum:
                      - foo
                      - bar
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} root:
        :{typedirective} root: string:enum
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["req_or_res", "directive", "typedirective"],
    [
        pytest.param("req", "reqjson", "reqjsonobj", id="req"),
        pytest.param("res", "resjson", "resjsonobj", id="res"),
    ],
)
def test_render_json_schema_description_enum_wo_type(
    testrenderer, oas_fragment, req_or_res, directive, typedirective
):
    """JSON schema description is generated for JSON enum wo/ type."""

    markup = textify(
        testrenderer.render_json_schema_description(
            oas_fragment(
                """
                type: object
                properties:
                  root:
                    enum:
                      - foo
                      - bar
                """
            ),
            req_or_res,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :{directive} root:
        :{typedirective} root: enum
        """.rstrip()
    )



"""OpenAPI spec renderer: render_response."""

import textwrap
import pytest

from sphinxcontrib.openapi import renderers


def textify(generator):
    return "\n".join(generator)


@pytest.mark.parametrize(
    ["statuscode"], [pytest.param("200"), pytest.param("4XX"), pytest.param("default")]
)
def test_render_response_status_code(testrenderer, oas_fragment, statuscode):
    """Path response's definition is rendered for any status code."""

    markup = textify(
        testrenderer.render_response(
            statuscode,
            oas_fragment(
                """
                description: An evidence.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode %s:
           An evidence.
        """.rstrip()
        % statuscode
    )


def test_render_response_minimal(testrenderer, oas_fragment):
    """Path response's minimal definition is rendered."""

    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.
        """.rstrip()
    )


def test_render_response_description_commonmark_default(testrenderer, oas_fragment):
    """Path response's 'description' must be in commonmark."""

    markup = textify(
        testrenderer.render_response(
            "200",
            oas_fragment(
                """
                description: |
                  An __evidence__ that matches
                  the `query`.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An **evidence** that matches
           the ``query``.
        """.rstrip()
    )


def test_render_response_description_commonmark(fakestate, oas_fragment):
    """Path response's 'description' can be in commonmark."""

    testrenderer = renderers.HttpdomainRenderer(fakestate, {"markup": "commonmark"})
    markup = textify(
        testrenderer.render_response(
            "200",
            oas_fragment(
                """
                description: |
                  An __evidence__ that matches
                  the `query`.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An **evidence** that matches
           the ``query``.
        """.rstrip()
    )


def test_render_response_description_restructuredtext(fakestate, oas_fragment):
    """Path response's 'description' can be in restructuredtext."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"markup": "restructuredtext"}
    )
    markup = textify(
        testrenderer.render_response(
            "200",
            oas_fragment(
                """
                description: |
                  An __evidence__ that matches
                  the `query`.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An __evidence__ that matches
           the `query`.
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["status_code", "status"],
    [
        pytest.param("200", "OK", id="200"),
        pytest.param("201", "Created", id="201"),
        pytest.param("202", "Accepted", id="202"),
    ],
)
def test_render_response_content_2xx(testrenderer, oas_fragment, status_code, status):
    """Path response's 'content' definition is rendered."""

    markup = textify(
        testrenderer.render_response(
            status_code,
            oas_fragment(
                """
                description: An evidence.
                content:
                  application/json:
                    example:
                      foo: bar
                      baz: 42
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :statuscode {status_code}:
           An evidence.

           .. sourcecode:: http

              HTTP/1.1 {status_code} {status}
              Content-Type: application/json

              {{
                "foo": "bar",
                "baz": 42
              }}
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["status_code"],
    [
        pytest.param("301"),
        pytest.param("307"),
        pytest.param("401"),
        pytest.param("422"),
        pytest.param("502"),
    ],
)
def test_render_response_content_non_2xx(testrenderer, oas_fragment, status_code):
    """Path response's 'content' definition is NOT rendered."""

    markup = textify(
        testrenderer.render_response(
            status_code,
            oas_fragment(
                """
                description: An evidence.
                content:
                  application/json:
                    example:
                      foo: bar
                      baz: 42
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :statuscode {status_code}:
           An evidence.
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["status_code", "status"],
    [
        pytest.param("301", "Moved Permanently", id="301"),
        pytest.param("307", "Temporary Redirect", id="307"),
        pytest.param("401", "Unauthorized", id="401"),
        pytest.param("422", "Unprocessable Entity", id="422"),
    ],
)
def test_render_response_content_custom(fakestate, oas_fragment, status_code, status):
    """Path response's 'content' definition is rendered."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"response-examples-for": ["301", "307", "401", "422"]}
    )

    markup = textify(
        testrenderer.render_response(
            status_code,
            oas_fragment(
                """
                description: An evidence.
                content:
                  application/json:
                    example:
                      foo: bar
                      baz: 42
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :statuscode {status_code}:
           An evidence.

           .. sourcecode:: http

              HTTP/1.1 {status_code} {status}
              Content-Type: application/json

              {{
                "foo": "bar",
                "baz": 42
              }}
        """.rstrip()
    )


def test_render_response_content_custom_mismatch(fakestate, oas_fragment):
    """Path response's 'content' definition is NOT rendered."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"response-examples-for": ["301", "307", "401", "422"]}
    )

    markup = textify(
        testrenderer.render_response(
            "200",
            oas_fragment(
                """
                description: An evidence.
                content:
                  application/json:
                    example:
                      foo: bar
                      baz: 42
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        :statuscode 200:
           An evidence.
        """.rstrip()
    )


def test_render_response_header(testrenderer, oas_fragment):
    """Path response's 'header' definition is rendered."""

    markup = textify(
        testrenderer.render_response(
            "200",
            oas_fragment(
                """
                description: An evidence.
                headers:
                  X-Request-Id:
                    description: A unique request identifier.
                    schema:
                      type: string
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        :resheader X-Request-Id:
           A unique request identifier.
        :resheadertype X-Request-Id: string
        """.rstrip()
    )


def test_render_response_header_minimal(testrenderer, oas_fragment):
    """Path response's 'header' minimal definition is rendered."""

    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                headers:
                  X-Request-Id: {}
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        :resheader X-Request-Id:
        """.rstrip()
    )


def test_render_response_header_description(testrenderer, oas_fragment):
    """Path response's 'header' description is rendered."""

    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                headers:
                  X-Request-Id:
                    description: A unique request identifier.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        :resheader X-Request-Id:
           A unique request identifier.
        """.rstrip()
    )


def test_render_response_header_multiline_description(testrenderer, oas_fragment):
    """Path response's 'header' multiline description is rendered."""

    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                headers:
                  X-Request-Id:
                    description: |
                      A unique request
                      identifier.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        :resheader X-Request-Id:
           A unique request
           identifier.
        """.rstrip()
    )


def test_render_response_header_description_commonmark_default(
    testrenderer, oas_fragment
):
    """Path response's 'header' description must be in commonmark by default."""

    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                headers:
                  X-Request-Id:
                    description: |
                      A unique __request__
                      `identifier`.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        :resheader X-Request-Id:
           A unique **request**
           ``identifier``.
        """.rstrip()
    )


def test_render_response_header_description_commonmark(fakestate, oas_fragment):
    """Path response's 'header' description can be in commonmark."""

    testrenderer = renderers.HttpdomainRenderer(fakestate, {"markup": "commonmark"})
    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                headers:
                  X-Request-Id:
                    description: |
                      A unique __request__
                      `identifier`.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        :resheader X-Request-Id:
           A unique **request**
           ``identifier``.
        """.rstrip()
    )


def test_render_response_header_description_restructuredtext(fakestate, oas_fragment):
    """Path response's 'header' description can be in restructuredtext."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"markup": "restructuredtext"}
    )
    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                headers:
                  X-Request-Id:
                    description: |
                      A unique __request__
                      `identifier`.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        :resheader X-Request-Id:
           A unique __request__
           `identifier`.
        """.rstrip()
    )


def test_render_response_header_content_type(testrenderer, oas_fragment):
    """Path response's 'Content-Type' header is ignored."""

    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                headers:
                  Content-Type: {}
                """
            ),
        )
    )

    # There's an extra newline at the end of markup if there's at least one
    # response header defined.
    assert markup.rstrip() == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        """.rstrip()
    )


def test_render_response_header_required(testrenderer, oas_fragment):
    """Path response's header 'required' marker is rendered."""

    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                headers:
                  X-Request-Id:
                    required: true
                """
            ),
        )
    )

    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        :resheader X-Request-Id:
        :resheadertype X-Request-Id: required
        """.rstrip()
    )


def test_render_response_header_required_false(testrenderer, oas_fragment):
    """Path response's header 'required' marker is not rendered."""

    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                headers:
                  X-Request-Id:
                    required: false
                """
            ),
        )
    )

    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        :resheader X-Request-Id:
        """.rstrip()
    )


def test_render_response_header_deprecated(testrenderer, oas_fragment):
    """Path response's header 'deprecated' marker is rendered."""

    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                headers:
                  X-Request-Id:
                    deprecated: true
                """
            ),
        )
    )

    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        :resheader X-Request-Id:
        :resheadertype X-Request-Id: deprecated
        """.rstrip()
    )


def test_render_response_header_deprecated_false(testrenderer, oas_fragment):
    """Path response's header 'deprecated' marker is not rendered."""

    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                headers:
                  X-Request-Id:
                    deprecated: false
                """
            ),
        )
    )

    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        :resheader X-Request-Id:
        """.rstrip()
    )


def test_render_response_header_required_deprecated(testrenderer, oas_fragment):
    """Path response's header markers are rendered."""

    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                headers:
                  X-Request-Id:
                    required: true
                    deprecated: true
                """
            ),
        )
    )

    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        :resheader X-Request-Id:
        :resheadertype X-Request-Id: required, deprecated
        """.rstrip()
    )


def test_render_response_header_type(testrenderer, oas_fragment):
    """Path response's header type is rendered."""

    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                headers:
                  X-Request-Id:
                    schema:
                      type: string
                """
            ),
        )
    )

    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        :resheader X-Request-Id:
        :resheadertype X-Request-Id: string
        """.rstrip()
    )


def test_render_response_header_type_with_format(testrenderer, oas_fragment):
    """Path response's header type with format is rendered."""

    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                headers:
                  X-Request-Id:
                    schema:
                      type: string
                      format: uuid4
                """
            ),
        )
    )

    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        :resheader X-Request-Id:
        :resheadertype X-Request-Id: string:uuid4
        """.rstrip()
    )


def test_render_response_header_type_from_content(testrenderer, oas_fragment):
    """Path response's header type from content is rendered."""

    markup = textify(
        testrenderer.render_response(
            200,
            oas_fragment(
                """
                description: An evidence.
                headers:
                  X-Request-Id:
                    content:
                      text/plain:
                        schema:
                          type: string
                """
            ),
        )
    )

    assert markup == textwrap.dedent(
        """\
        :statuscode 200:
           An evidence.

        :resheader X-Request-Id:
        :resheadertype X-Request-Id: string
        """.rstrip()
    )



import os
import pathlib
import textwrap

import pytest
import yaml

from sphinx.application import Sphinx
from sphinxcontrib.openapi import utils


_testspecs_dir = pathlib.Path(os.path.dirname(__file__), "testspecs")
_testspecs_v2_dir = _testspecs_dir.joinpath("v2.0")
_testspecs_v3_dir = _testspecs_dir.joinpath("v3.0")

_testspecs = [str(path.relative_to(_testspecs_dir)) for path in _testspecs_dir.glob("*/*")]
_testspecs_v2 = [str(path.relative_to(_testspecs_dir)) for path in _testspecs_v2_dir.glob("*/*")]
_testspecs_v3 = [str(path.relative_to(_testspecs_dir)) for path in _testspecs_v3_dir.glob("*/*")]


def pytest_addoption(parser):
    parser.addoption("--regenerate-rendered-specs", action="store_true")


def pytest_collection_modifyitems(items):
    items_new = []

    for item in items:
        has_mark = bool(list(item.iter_markers(name="regenerate_rendered_specs")))

        if any(
            [
                item.config.getoption("--regenerate-rendered-specs") and has_mark,
                not item.config.getoption("--regenerate-rendered-specs") and not has_mark,
            ]
        ):
            items_new.append(item)

    items[:] = items_new


def _format_option_raw(key, val):
    if isinstance(val, bool) and val:
        return ':%s:' % key
    return ':%s: %s' % (key, val)


@pytest.fixture(scope='function')
def run_sphinx(tmpdir):
    src = tmpdir.ensure('src', dir=True)
    out = tmpdir.ensure('out', dir=True)

    def run(spec, options={}):
        options_raw = '\n'.join([
            '   %s' % _format_option_raw(key, val)
            for key, val in options.items()])

        src.join('conf.py').write_text(
            textwrap.dedent('''
                import os

                project = 'sphinxcontrib-openapi-test'
                copyright = '2017, Ihor Kalnytskyi'

                extensions = ['sphinxcontrib.openapi']
                source_suffix = '.rst'
                master_doc = 'index'
            '''),
            encoding='utf-8')

        src.join('index.rst').write_text(
            '.. openapi:: %s\n%s' % (spec, options_raw),
            encoding='utf-8')

        Sphinx(
            srcdir=src.strpath,
            confdir=src.strpath,
            outdir=out.strpath,
            doctreedir=out.join('.doctrees').strpath,
            buildername='html'
        ).build()

    yield run


@pytest.fixture(scope="function")
def get_testspec():
    def get_testspec(*args, encoding="utf-8", resolve_refs=True):
        with _testspecs_dir.joinpath(*args).open(encoding=encoding) as f:
            spec = yaml.safe_load(f)
            if resolve_refs:
                spec = utils._resolve_refs("", spec)
            return spec
    return get_testspec


@pytest.fixture(scope="function", params=_testspecs)
def testspec(request, get_testspec):
    return request.param, get_testspec(request.param)



"""
    sphinxcontrib.openapi
    ---------------------

    The OpenAPI spec renderer for Sphinx. It's a new way to document your
    RESTful API. Based on ``sphinxcontrib-httpdomain``.

    :copyright: (c) 2016, Ihor Kalnytskyi.
    :license: BSD, see LICENSE for details.
"""

import os
from pkg_resources import get_distribution, DistributionNotFound
from sphinxcontrib.openapi import renderers, directive
from sphinx.domains import Domain
import yaml
from docutils import nodes

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    __version__ = None


_DEFAULT_RENDERER_NAME = "httpdomain:old"


def oasversion_role(typ, rawtext, text, lineno, inliner, options={}, content=[]):
    fn = text
    env = inliner.document.settings.env
    rel_fn, fn = env.relfn2path(fn)
    y = yaml.load(open(fn, 'r').read(), Loader=yaml.FullLoader)
    s = y['info']['version']
    retnode = nodes.inline(text=s, role=typ.lower(), classes=[typ])
    return [retnode], []


class OpenAPIDomain(Domain):
    name = 'openapi'
    label = 'OpenAPI Documentation'

    directives = {
        'httpdomain:old': directive.create_directive_from_renderer(
            renderers.HttpdomainOldRenderer),
        'httpdomain': directive.create_directive_from_renderer(renderers.HttpdomainRenderer),
        'model': directive.create_directive_from_renderer(renderers.ModelRenderer),
        'toc': directive.create_directive_from_renderer(renderers.TocRenderer),
    }

    roles = {
        'version': oasversion_role
    }


def setup(app):
    app.add_config_value("openapi_default_renderer", _DEFAULT_RENDERER_NAME, "html")
    app.add_config_value("openapi_renderers", {}, "html")

    package_dir = os.path.abspath(os.path.dirname(__file__))
    locale_dir = os.path.join(package_dir, 'locale')
    app.add_message_catalog('openapi', locale_dir)

    from sphinxcontrib import httpdomain

    for idx, fieldtype in enumerate(httpdomain.HTTPResource.doc_field_types):
        if fieldtype.name == 'requestheader':
            httpdomain.HTTPResource.doc_field_types[idx] = httpdomain.TypedField(
                fieldtype.name,
                label=fieldtype.label,
                names=fieldtype.names,
                typerolename='header',
                typenames=('reqheadertype', ),
            )

        if fieldtype.name == 'responseheader':
            httpdomain.HTTPResource.doc_field_types[idx] = httpdomain.TypedField(
                fieldtype.name,
                label=fieldtype.label,
                names=fieldtype.names,
                typerolename='header',
                typenames=('resheadertype', ),
            )

    app.setup_extension("sphinxcontrib.httpdomain")
    app.add_domain(OpenAPIDomain)
    app.add_directive(
        "openapi",
        directive.create_directive_from_renderer(renderers.HttpdomainOldRenderer)
    )

    return {"version": __version__, "parallel_read_safe": True}



import pytest

from sphinxcontrib.openapi.schema_utils import example_from_schema


@pytest.mark.parametrize(
    ["schema", "expected"],
    [
        pytest.param(
            {
                "type": "object",
                "properties": {
                    "string": {"type": "string"},
                    "integer": {"type": "integer"},
                    "number": {"type": "number"},
                    "array_of_numbers": {"type": "array", "items": {"type": "number"}},
                    "array_of_strings": {"type": "array", "items": {"type": "string"}},
                },
            },
            {
                "string": "string",
                "integer": 1,
                "number": 1.0,
                "array_of_numbers": [1.0, 1.0],
                "array_of_strings": ["string", "string"],
            },
            id="no_examples",
        ),
        pytest.param(
            {
                "type": "object",
                "properties": {
                    "string": {"type": "string", "example": "example_string"},
                    "integer": {"type": "integer", "example": 2},
                    "number": {"type": "number", "example": 2.0},
                    "array_of_numbers": {
                        "type": "array",
                        "items": {"type": "number", "example": 3.0},
                    },
                    "array_of_strings": {
                        "type": "array",
                        "items": {"type": "string", "example": "example_string"},
                    },
                },
            },
            {
                "string": "example_string",
                "integer": 2,
                "number": 2.0,
                "array_of_numbers": [3.0, 3.0],
                "array_of_strings": ["example_string", "example_string"],
            },
            id="with_examples",
        ),
        pytest.param(
            {
                "type": "object",
                "properties": {
                    "string": {"type": "string"},
                    "date": {"type": "string", "format": "date"},
                    "date-time": {"type": "string", "format": "date-time"},
                    "password": {"type": "string", "format": "password"},
                    "byte": {"type": "string", "format": "byte"},
                    "ipv4": {"type": "string", "format": "ipv4"},
                    "ipv6": {"type": "string", "format": "ipv6"},
                    "unknown": {"type": "string", "format": "unknown"},
                    "max": {"type": "string", "maxLength": 3},
                    "min": {"type": "string", "minLength": 10},
                    "max_min_big": {"type": "string", "maxLength": 15, "minLength": 10},
                    "max_min_small": {"type": "string", "maxLength": 3, "minLength": 1},
                },
            },
            {
                "string": "string",
                "date": "2020-01-01",
                "date-time": "2020-01-01T01:01:01Z",
                "password": "********",
                "byte": "QG1pY2hhZWxncmFoYW1ldmFucw==",
                "ipv4": "127.0.0.1",
                "ipv6": "::1",
                "unknown": "string",
                "max": "str",
                "min": "stringstri",
                "max_min_big": "stringstri",
                "max_min_small": "str",
            },
            id="strings",
        ),
        pytest.param(
            {
                "type": "object",
                "properties": {
                    "inner_object": {"type": "object"},
                    "inner_array_of_objects": {
                        "type": "array",
                        "items": {"type": "object"},
                    },
                    "inner_array_of_arrays": {
                        "type": "array",
                        "items": {"type": "array", "items": {"type": "integer"}},
                    },
                },
            },
            {
                "inner_object": {},
                "inner_array_of_objects": [{}, {}],
                "inner_array_of_arrays": [[1, 1], [1, 1]],
            },
            id="nested",
        ),
        pytest.param(
            {
                "type": "object",
                "properties": {
                    "string_enum": {"type": "string", "enum": ["three", "two", "one"]},
                    "integer_enum": {"type": "integer", "enum": [3, 2, 1]},
                },
            },
            {"string_enum": "three", "integer_enum": 3},
            id="enum",
        ),
        pytest.param(
            {
                "type": "object",
                "properties": {
                    "mixed_type_array": {
                        "type": "array",
                        "items": {"oneOf": {"string", "integer"}},
                    },
                    "any_type_array": {"type": "array", "items": {}},
                    "0_to_1_length_array": {
                        "type": "array",
                        "minItems": 0,
                        "maxItems": 1,
                        "items": {"type": "integer"},
                    },
                    "5_to_10_length_array": {
                        "type": "array",
                        "minItems": 5,
                        "maxItems": 10,
                        "items": {"type": "integer"},
                    },
                },
            },
            {
                "mixed_type_array": [1, 1],
                "any_type_array": ["string", 1],
                "0_to_1_length_array": [1],
                "5_to_10_length_array": [1, 1, 1, 1, 1],
            },
            id="arrays",
        ),
        pytest.param(
            {
                "type": "object",
                "properties": {
                    "oneOf": {
                        "oneOf": [
                            {
                                "type": "object",
                                "properties": {"one": {"type": "string"}},
                            },
                            {
                                "type": "object",
                                "properties": {"two": {"type": "string"}},
                            },
                        ]
                    },
                    "anyOf": {
                        "anyOf": [
                            {
                                "type": "object",
                                "properties": {"three": {"type": "string"}},
                            },
                            {
                                "type": "object",
                                "properties": {"four": {"type": "string"}},
                            },
                        ]
                    },
                    "allOf": {
                        "allOf": [
                            {
                                "type": "object",
                                "properties": {"five": {"type": "string"}},
                            },
                            {
                                "type": "object",
                                "properties": {"six": {"type": "string"}},
                            },
                        ]
                    },
                },
            },
            {
                "oneOf": {"one": "string"},
                "anyOf": {"three": "string"},
                "allOf": {"five": "string", "six": "string"},
            },
            id="oneOf_anyOf_allOf",
        ),
        pytest.param(
            {
                "type": "object",
                "properties": {"anything": {"description": "this can be anything"}},
            },
            {"anything": 1},
            id="any_type",
        ),
        pytest.param(
            {
                "type": "object",
                "properties": {
                    "min_int": {"type": "integer", "minimum": 10},
                    "max_int": {"type": "integer", "maximum": -10},
                    "min_and_max_int": {
                        "type": "integer",
                        "minimum": 10,
                        "maximum": 20,
                    },
                    "exclusive_int": {
                        "type": "integer",
                        "minimum": -10,
                        "maximum": 10,
                        "exclusiveMinimum": True,
                        "exclusiveMaximum": True,
                    },
                    "min_num": {"type": "number", "minimum": 10.0},
                    "max_num": {"type": "number", "maximum": -10.0},
                    "exclusive_num": {
                        "type": "number",
                        "minimum": -10,
                        "maximum": 10,
                        "exclusiveMinimum": True,
                        "exclusiveMaximum": True,
                    },
                },
            },
            {
                "min_int": 11,
                "max_int": -11,
                "min_and_max_int": 15,
                "exclusive_int": 0,
                "min_num": 11.0,
                "max_num": -11.0,
                "exclusive_num": 0.0,
            },
            id="min_max",
        ),
    ],
)
def test_generate_example_from_schema(schema, expected):
    assert example_from_schema(schema) == expected
